import matplotlib.pyplot as plt
import seaborn as sns
import pandas as pd
import numpy as np
from os.path import exists, basename,dirname
from icecream import ic as info
from roux.viz.ax_ import *

def plot_range(df00: pd.DataFrame,
               colvalue: str,
               colindex: str,
               k: str,
               headsize: int=15,
               headcolor: str='lightgray',
    ax: plt.Axes=None,
    ) -> plt.Axes:
    """Plot range/intervals e.g. genome coordinates as lines.

    Args:
        df00 (pd.DataFrame): input data.
        colvalue (str): column with values.
        colindex (str): column with ids.
        k (str): subset name.
        headsize (int, optional): margin at top. Defaults to 15.
        headcolor (str, optional): color of the margin. Defaults to 'lightgray'.
        ax (plt.Axes, optional): `plt.Axes` object. Defaults to None.

    Returns:
        plt.Axes: `plt.Axes` object.
    """
    df00['rank']=df00[colvalue].rank()
    x,y=df00.rd.filter_rows({colindex:k}).iloc[0,:]['rank'],df00.rd.filter_rows({colindex:k}).iloc[0,:][colvalue]
    if ax is None:
        fig,ax=plt.subplots(figsize=[1,1])
    ax=df00.set_index('rank').sort_index(0)[colvalue].plot.area(ax=ax)
    ax.annotate('', xy=(x, y),  xycoords='data',
                xytext=(x, ax.get_ylim()[1]), textcoords='data',
                arrowprops=dict(facecolor=headcolor, shrink=0,
                               width=0,ec='none',
                               headwidth=headsize,
                               headlength=headsize,
                               ),
                horizontalalignment='right', verticalalignment='top',
                )
    d_=get_axlims(ax)
    ax.text(x,y+(d_['y']['len'])*0.25,int(y),#f"{y:.1f}",
                    # transform=ax.transAxes,
                    va='bottom',ha='center',
                   )
    ax.text(0.5,0,colvalue,
                    transform=ax.transAxes,
                    va='top',ha='center',
                   )
    ax.axis(False)
    return ax
    
def plot_connections(
    dplot: pd.DataFrame,
    label2xy: dict,
    colval: str='$r_{s}$',
    line_scale: int=40,
    legend_title: str='similarity',
    label2rename: dict=None,
    element2color: dict=None,
    xoff: float=0,
    yoff: float=0,
    rectangle: dict={'width':0.2,'height':0.32},
    params_text: dict={'ha':'center','va':'center'},
    params_legend: dict={'bbox_to_anchor':(1.1, 0.5),
    'ncol':1,
    'frameon':False},
    legend_elements: list=[],
    params_line: dict={'alpha':1},
    ax: plt.Axes=None,
    test: bool=False
    ) -> plt.Axes:
    """Plot connections between points with annotations.

    Args:
        dplot (pd.DataFrame): input data.
        label2xy (dict): label to position.
        colval (str, optional): column with values. Defaults to '{s}$'.
        line_scale (int, optional): line_scale. Defaults to 40.
        legend_title (str, optional): legend_title. Defaults to 'similarity'.
        label2rename (dict, optional): label2rename. Defaults to None.
        element2color (dict, optional): element2color. Defaults to None.
        xoff (float, optional): xoff. Defaults to 0.
        yoff (float, optional): yoff. Defaults to 0.
        rectangle (_type_, optional): rectangle. Defaults to {'width':0.2,'height':0.32}.
        params_text (_type_, optional): params_text. Defaults to {'ha':'center','va':'center'}.
        params_legend (_type_, optional): params_legend. Defaults to {'bbox_to_anchor':(1.1, 0.5), 'ncol':1, 'frameon':False}.
        legend_elements (list, optional): legend_elements. Defaults to [].
        params_line (_type_, optional): params_line. Defaults to {'alpha':1}.
        ax (plt.Axes, optional): `plt.Axes` object. Defaults to None.
        test (bool, optional): test mode. Defaults to False.

    Returns:
        plt.Axes: `plt.Axes` object.
    """
    import matplotlib.patches as mpatches
    from matplotlib.collections import PatchCollection
    label2xy={k:[label2xy[k][0]+xoff,label2xy[k][1]+yoff] for k in label2xy}
    dplot['index xy']=dplot['index'].map(label2xy)
    dplot['column xy']=dplot['column'].map(label2xy)
    
    ax=plt.subplot() if ax is None else ax
    from roux.viz.ax_ import set_logos,get_subplot_dimentions
    patches=[]
    label2xys_rectangle_centers={}
    for label in label2xy:
        xy=label2xy[label]
        rect = mpatches.Rectangle(xy, **rectangle, fill=False,fc="none",lw=2,
                                  ec=element2color[label] if not element2color is None else 'k',
                                 zorder=0)

        patches.append(rect)
        line_xys=[np.transpose(np.array(rect.get_bbox()))[0],np.transpose(np.array(rect.get_bbox()))[1][::-1]]
        label2xys_rectangle_centers[label]=[np.mean(line_xys[0]),np.mean(line_xys[1])]
        inset_width=0.2
        inset_height=inset_width/get_subplot_dimentions(ax)[2]
        axin=ax.inset_axes([*[l-(off*0.5) for l,off in zip(label2xys_rectangle_centers[label],[inset_width,inset_height])],
                            inset_width,inset_height])
        if not test:
            axin=set_logos(label=label,element2color=element2color,ax=axin,test=test)
        axin.text(np.mean(axin.get_xlim()),np.mean(axin.get_ylim()),
                 label2rename[label] if not label2rename is None else label,
                  **params_text,
                 )
    dplot.apply(lambda x: ax.plot(*[[label2xys_rectangle_centers[x[k]][0] for k in ['index','column']],
                                  [label2xys_rectangle_centers[x[k]][1] for k in ['index','column']]],
                                  lw=(x[colval]-0.49)*line_scale,
                                  linestyle=params_line['linestyle'],
                                  color='k',zorder=-1,
                                  alpha=params_line['alpha'],
                                ),axis=1)            
    if params_line['annot']:
        def set_text_position(ax,x):
            xs,ys=[[label2xys_rectangle_centers[x[k]][i] for k in ['index','column']] for i in [0,1]]
            xy=[np.mean(xs),np.mean(ys)]
            if np.subtract(*xs)==0 or np.subtract(*ys)==0:
                ha,va='center','center'
                rotation=0
            else:
                if np.subtract(*xs)<0:      
                    ha,va='right','bottom'
                    xy[1]=xy[1]+0.025
                    rotation=-45
                else:
                    ha,va='right','top'
                    xy[1]=xy[1]-0.025
                    rotation=45
            ax.text(xy[0],xy[1],f"{x[colval]:.2f}",
                    ha=ha,va=va,
                    color='k',rotation=rotation,
                   bbox=dict(boxstyle="round",
                   fc='lightgray',ec=None,)
                   )
            return ax
            
        dplot.apply(lambda x: set_text_position(ax,x),axis=1)            
    from matplotlib.lines import Line2D
    legend_elements=legend_elements+[Line2D([0], [0], color='k', linestyle='solid', lw=(i-0.49)*line_scale, 
                                alpha=params_line['alpha'],
                                label=f' {colval}={i:1.1f}') for i in [1.0,0.8,0.6]]
    ax.legend(handles=legend_elements,
              title=legend_title,**params_legend)
    ax.set(**{'xlim':[0,1],'ylim':[0,1]})
    if not test:
        ax.set_axis_off()      
    return ax

def plot_kinetics(
    df1: pd.DataFrame,
    x: str,
    y: str,
    hue: str,
    cmap: str='Reds_r',
    ax: plt.Axes=None,
    test: bool=False,
    kws_legend: dict={},
    **kws_set,
    ) -> plt.Axes:
    """Plot time-dependent kinetic data.

    Args:
        df1 (pd.DataFrame): input data.
        x (str): x column.
        y (str): y column.
        hue (str): hue column.
        cmap (str, optional): colormap. Defaults to 'Reds_r'.
        ax (plt.Axes, optional): `plt.Axes` object. Defaults to None.
        test (bool, optional): test mode. Defaults to False.
        kws_legend (dict, optional): legend parameters. Defaults to {}.

    Returns:
        plt.Axes: `plt.Axes` object.
    """
    from roux.viz.ax_ import rename_legends
    from roux.viz.colors import get_ncolors
    df1=df1.sort_values(hue,ascending=False)
    info(df1[hue].unique())
    if ax is None: fig,ax=plt.subplots(figsize=[2.5,2.5])
    label2color=dict(zip(df1[hue].unique(),get_ncolors(df1[hue].nunique(),
                                                            ceil=False,
                                                            cmap=cmap,
                                                                )))
    df2=df1.groupby([hue,x],sort=False).agg({c:[np.mean,np.std] for c in [y]}).rd.flatten_columns().reset_index()
    d1=df1.groupby([hue,x],sort=False,as_index=False).size().groupby(hue)['size'].agg([min,max]).T.to_dict()
    d2={str(k):str(k)+'\n'+(f"(n={d1[k]['min']})" if d1[k]['min']==d1[k]['max'] else f"(n={d1[k]['min']}-{d1[k]['max']})") for k in d1}
    if test:info(d2)
    df2.groupby(hue,sort=False).apply(lambda df: df.sort_values(x).plot(x=x,
                                                            y=f"{y} mean",
                                                            yerr=f"{y} std",
                                                                        elinewidth=0.3,
                                                            label=df.name,
                                                            color=label2color[df.name],
                                                            lw=2,
                                                           ax=ax))
    ax=rename_legends(ax,replaces=d2,title=hue,
                     **kws_legend)
    ax.set(**kws_set)
    return ax


## plot data shape changes
def plot_steps(df1,
               coln='n',
               ax=None,
               test=False):
    """
    changes in numbers
    """
    df1=df1.sort_values(coln,ascending=False)
    df1['% change']=df1[coln].pct_change() * 100
    df1['y']=range(len(df1))
    if ax is None:
        fig,ax=plt.subplots(figsize=[4,len(df1)])
    kws_line=dict( marker='o',mfc='w',color='gray',ms=17)
    df1.iloc[:-1,:].apply(lambda x: ax.plot([0,0],[x['y'],x['y']+1],**kws_line),axis=1)
    df1.apply(lambda x: ax.text(0.005,x['y'],s=x.name,ha='left',va='center'),axis=1)
    df1.apply(lambda x: ax.text(0,x['y'],s=f"{x['y']:.0f}",ha='center',va='center'),axis=1)
    df1.apply(lambda x: ax.text(0.005,x['y']+0.33,s=f"n={x[coln]:.0f}",ha='left',va='center',
                               alpha=0.75),axis=1)
    from rohan.lib.plot.colors import saturate_color
    df1.apply(lambda x: ax.text(0.005,x['y']+0.66,s=f"{x['% change']:.1f}%" if not pd.isnull(x['% change']) else '',ha='left',va='center',
                               color=saturate_color("#FF0000",0.5+(x['% change']/-100)) if x['% change']<0 else 'g',
                               alpha=0.75),axis=1)
    ax.set(xlim=[-0.005,0.1],
          ylim=[len(df1)+0.5,-1.5])
    if not test:ax.axis('off')
    return ax


# from roux.global_imports import *
import matplotlib.pyplot as plt
from roux.lib.io import *#read_ps,to_outp
from roux.lib.sys import is_interactive_notebook

## matplotlib plots
def to_plotp(
    ax: plt.Axes,
    prefix: str='plot/plot_',
    suffix: str='',
    fmts: list=['png'],
    ) -> str:
    """Infer output path for a plot.

    Args:
        ax (plt.Axes): `plt.Axes` object.
        prefix (str, optional): prefix with directory path for the plot. Defaults to 'plot/plot_'.
        suffix (str, optional): suffix of the filename. Defaults to ''.
        fmts (list, optional): formats of the images. Defaults to ['png'].

    Returns:
        str: output path for the plot.
    """
    if isinstance(ax,str):
        return ax
    if isinstance(suffix,(list,tuple)):
        suffix='_'.join(suffix)
    suffix=suffix.replace('/','_')
    plotp=prefix
    for k in ['get_xlabel','get_ylabel','get_title','legend_']: 
        if hasattr(ax,k):
            if k!='legend_':
                plotp=f"{plotp}_"+(getattr(ax,k)()).replace('.','_')
            else:
                if not ax.legend_ is None:
                    plotp=f"{plotp}_"+(ax.legend_.get_title().get_text()).replace('.','_')
    plotp=f"{plotp} {suffix}"+(f".{fmts[0]}" if len(fmts)==1 else '')
    return plotp

def savefig(
    plotp: str,
    tight_layout: bool=True,
    bbox_inches: list=None, # overrides tight_layout
    fmts: list=['png'],
    savepdf: bool=False,
    normalise_path: bool=True,
    replaces_plotp: dict=None,
    dpi: int=500,
    force: bool=True,
    kws_replace_many: dict={},
    kws_savefig: dict={},
    **kws,
    ) -> str:
    """Wrapper around `plt.savefig`.

    Args:
        plotp (str): output path or `plt.Axes` object.
        tight_layout (bool, optional): tight_layout. Defaults to True.
        bbox_inches (list, optional): bbox_inches. Defaults to None.
        savepdf (bool, optional): savepdf. Defaults to False.
        normalise_path (bool, optional): normalise_path. Defaults to True.
        replaces_plotp (dict, optional): replaces_plotp. Defaults to None.
        dpi (int, optional): dpi. Defaults to 500.
        force (bool, optional): overwrite output. Defaults to True.
        kws_replace_many (dict, optional): parameters provided to the `replace_many` function. Defaults to {}.
    
    Keyword Args:
        kws: parameters provided to `to_plotp` function.
        kws_savefig: parameters provided to `to_savefig` function.
        kws_replace_many: parameters provided to `replace_many` function.
    
    Returns:
        str: output path.
    """
#         from roux.viz.ax_ import to_plotp
    plotp=to_plotp(plotp,
                   fmts=fmts,**kws)
    if not replaces_plotp is None:
        plotp=replace_many(plotp,replaces=replaces_plotp,)
    if exists(plotp):
        logging.warning(f"overwritting: {plotp}")
    if plotp.count('.')>1:
        plotp=abspath(plotp)        
        if plotp.count('.')>1:
            logging.error(f"more than one '.' not allowed in the path {plotp}")
            return 
    if normalise_path:
        plotp=abspath(to_path(plotp))
    plotp=f"{dirname(plotp)}/{basenamenoext(plotp).replace('.','_')}{splitext(plotp)[1]}"    
    makedirs(plotp,exist_ok=True)
    if len(fmts)==0:
        fmts=['png']
    if exists(plotp) and not force:
        logging.info("fig exists")
        return
    if '.' in plotp:
        plt.savefig(plotp,
                    dpi=dpi,
                    bbox_inches=bbox_inches if (not bbox_inches is None) else 'tight' if tight_layout else None,
                    **kws_savefig
                   )
    else:
        for fmt in fmts:
            plotp=f"{plotp}.{fmt}"
            if exists(plotp) and not force:
                logging.info("fig exists")
                return
            plt.savefig(plotp,
                        format=fmt,
                        dpi=dpi,
                        bbox_inches=bbox_inches if (not bbox_inches is None) else 'tight' if tight_layout else None,
                       **kws_savefig,
                       )
    if not is_interactive_notebook():
        plt.clf();plt.close()
    return plotp

def savelegend(
    plotp: str,
    legend: object,
    expand: list=[-5,-5,5,5],
    **kws_savefig
    ) -> str:
    """Save only the legend of the plot/figure.

    Args:
        plotp (str): output path.
        legend (object): legend object.
        expand (list, optional): expand. Defaults to [-5,-5,5,5].

    Returns:
        str: output path.

    References: 
        1. https://stackoverflow.com/a/47749903/3521099
    """
    fig  = legend.figure
    fig.canvas.draw()
    bbox  = legend.get_window_extent()
    bbox = bbox.from_extents(*(bbox.extents + np.array(expand)))
    bbox = bbox.transformed(fig.dpi_scale_trans.inverted())
    plt.axis('off')
    return savefig(plotp,bbox_inches=bbox,
           **kws_savefig)

def update_kws_plot(
    kws_plot: dict,
    kws_plotp: dict,
    test: bool=False
    ) -> dict:
    """Update the input parameters.

    Args:
        kws_plot (dict): input parameters.
        kws_plotp (dict): saved parameters.
        test (bool, optional): _description_. Defaults to False.

    Returns:
        dict: updated parameters.
    """
    if exists(kws_plotp):
        kws_plot_=read_dict(kws_plotp)
    else:
        logging.error(f'not found{kws_plotp}')
        return
#     kws_plot=kws_plot_saved if kws_plot is None else {k:param[k] if k in kws_plot else kws_plot_saved[k] for k in kws_plot_saved};
    if test:print(kws_plot,kws_plot_)
    if kws_plot is None:
        return kws_plot_
    for k in kws_plot_:
        if not k in kws_plot:
            kws_plot[k]=kws_plot_[k]
    return kws_plot

def get_plot_inputs(
    plotp: str,
    df1: pd.DataFrame,
    kws_plot: dict,
    outd: str,
    ) -> tuple:
    """Get plot inputs.

    Args:
        plotp (str): path of the plot.
        df1 (pd.DataFrame): data for the plot.
        kws_plot (dict): parameters of the plot.
        outd (str): output directory.

    Returns:
        tuple: (path,dataframe,dict)
    """
    if exists(plotp):
        plotp=abspath(plotp)
    else:
        if not outd is None:
            plotp=f"{outd}/{plotp}"
            assert exists(plotp), f"not found {plotp}"
    if not outd is None:
        outd=abspath(outd)
        if not outd in plotp:
            plotp=f"{outd}/{plotp}"
            assert exists(plotp), f"not found {plotp}"
    else:
        ## remove suffixes
        outd=remove_exts(plotp)+'/'
    if df1 is None:
        df1=read_table(f"{outd}/df1.tsv");
    kws_plot=update_kws_plot(kws_plot,kws_plotp=f"{outd}/kws_plot.json")
    return plotp,df1,kws_plot

def log_code():
    """Log the code.
    """
    if is_interactive_notebook():
        logp=f'log_notebook.log'
        # if exists(logp):
            # open(logp, 'w').close()
            # logging.info(f'{logp} emptied')
        get_ipython().run_line_magic('logstart',f'{logp} over')        
        return 

def get_lines(
    logp: str='log_notebook.log',
    sep: str='# plot',
    test: bool=False
    ) -> list:
    """Get lines from the log.

    Args:
        logp (str, optional): path to the log file. Defaults to 'log_notebook.log'.
        sep (str, optional): label marking the start of code of the plot. Defaults to '# plot'.
        test (bool, optional): test mode. Defaults to False.

    Returns:
        list: lines of code. 
    """
    from roux.lib.text import read_lines 
    if not is_interactive_notebook():
        from roux.lib.sys import get_excecution_location
        p,i=get_excecution_location(depth=4)
        if test:print(p,i)
        l1=read_lines(p)[:i]
        if test:print(l1)
    else:
        if not exists(logp):
            log_code()
            logging.error(f'rerun the cell, {logp} created');
            return
        l1=read_lines(logp)
        if test: print(l1)
        # open(logp, 'w').close()
        # logging.info(f'{logp} emptied')
    if len(l1)==0: 
        logging.error('log code not found.');
        if is_interactive_notebook():
            log_code()
            logging.error(f'rerun the cell, because code not found');
        return
    lines=[]
    for linei,line in enumerate(l1[::-1]):
        line_=line.lstrip()
        if len(lines)==0 and ((not line_.startswith(f"to_plot(")) and (not line_.startswith(f"saveplot("))):
            continue
        if len(lines)==0:
            spaces=(len(line) - len(line.lstrip(' ')))#*' '
        line=line[spaces:]
        lines.append(line)
        if test:
            print(f"'{line}'")
        if any([line_.startswith(f"{sep} "),
                line_==f"{sep}\n",
                line_==f"{sep} \n"]):
            break
    lines=lines[::-1][1:-1]
    if len(lines)==0: logging.error('plot code not found.')
    return lines

def to_script(
    srcp: str,
    plotp: str,
    defn: str="plot_",
    s4: str='    ',
    test: bool=False,
    **kws
    ) -> str:
    """Save the script with the code for the plot.

    Args:
        srcp (str): path of the script.
        plotp (str): path of the plot.
        defn (str, optional): prefix of the function. Defaults to "plot_".
        s4 (str, optional): a tab. Defaults to '    '.
        test (bool, optional): test mode. Defaults to False.

    Returns:
        str: path of the script.
        
    TODOs:
        1. Compatible with names of the input dataframes other that `df1`.
            1. Get the variable name of the dataframe
            
            def get_df_name(df):
                name =[x for x in globals() if globals()[x] is df and not x.startswith('-')][0]
                return name
            
            2. Replace `df1` with the variable name of the dataframe.
    """
    lines=get_lines(**kws)
    if lines is None: return
    #make def
    for linei,line in enumerate(lines):
        if 'plt.subplot(' in line:
            lines[linei]=f'if ax is None:{line}'        
        if 'plt.subplots(' in line:
            lines[linei]=f'if ax is None:{line}'                
    lines=[f"    {l}" for l in lines]
    lines=''.join(lines)
    lines=f'def {defn}(\n{s4}plotp="{plotp}",\n{s4}df1=None,\n{s4}kws_plot=None,\n{s4}ax=None,\n{s4}fig=None,\n{s4}outd=None,\n{s4}fun_df1=None,\n{s4}**kws_set,\n{s4}):\n{s4}plotp,df1,kws_plot=get_plot_inputs(plotp=plotp,df1=df1,kws_plot=kws_plot,outd=f"{{dirname(__file__)}}");\n{s4}df1=fun_df1(df1) if not fun_df1 is None else df1;\n'+lines+f'{s4}ax.set(**kws_set)\n{s4}return ax\n'
    #save def
    with open(srcp,'w') as f:
        f.write('from roux.global_imports import *\n')
        f.write(lines)
    if test: print(lines)
    return srcp

def to_plot(
    plotp: str,
    df1: pd.DataFrame=None,
    kws_plot: dict=dict(),
    logp: str='log_notebook.log',
    sep: str='# plot',
    validate: bool=False,
    show_path: bool=False,
    show_path_offy: float=-0.2,
    force: bool=True,
    test: bool=False,
    quiet:bool=True,
    **kws) -> str:
    """Save a plot.

    Args:
        plotp (str): output path.
        df1 (pd.DataFrame, optional): dataframe with plotting data. Defaults to None.
        kws_plot (dict, optional): parameters for plotting. Defaults to dict().
        logp (str, optional): path to the log. Defaults to 'log_notebook.log'.
        sep (str, optional): separator marking the start of the plotting code in jupyter notebook. Defaults to '# plot'.
        validate (bool, optional): validate the "readability" using `read_plot` function. Defaults to False.
        show_path (bool, optional): show path on the plot. Defaults to False.
        show_path_offy (float, optional): y-offset for the path label. Defaults to 0.
        force (bool, optional): overwrite output. Defaults to True.
        test (bool, optional): test mode. Defaults to False.
        quiet (bool, optional): quiet mode. Defaults to False.
        
    Returns:
        str: output path.
    """
    #save plot
    plotp=savefig(plotp,force=force,**kws)
    if show_path:
        plt.figtext(x=0.5,y=0+show_path_offy,
                    s=plotp.split('plot/')[1] if 'plot/' in plotp else plotp,
                    ha='center')
    if df1 is None:
        if not quiet:
            logging.warning("no data provided to_plot")
        return plotp
    outd=plotp
    outd=remove_exts(outd)
    if test:
        info(outd)
    df1p=f"{outd}/df1.tsv"
    paramp=f"{outd}/kws_plot.json"    
    srcp=f"{outd}/plot.py"
    #save data
    to_table(df1,df1p)
    srcp=to_script(srcp=srcp,plotp=plotp,
                   logp=logp,sep=sep,test=test)
    if srcp is None: return plotp
    to_dict(kws_plot,paramp)
    if test:
        print({'plot':plotp,'data':df1p,'param':paramp})
    if validate:
        read_plot(srcp)
    return plotp

def read_plot(
    p: str,
    safe: bool=False,
    test: bool= False,
    **kws
    ) -> plt.Axes:
    """Generate the plot from data, parameters and a script.

    Args:
        p (str): path of the plot saved using `to_plot` function.
        safe (bool, optional): read as an image. Defaults to False.
        test (bool, optional): test mode. Defaults to False.
        
    Returns:
        plt.Axes: `plt.Axes` object.
    """
    if not safe:
        if not p.endswith('.py'):
            ## remove suffixes
            p=f"{remove_exts(p)}/plot.py"
            if test:info(p)
        from roux.workflow.io import import_from_file
        ax=import_from_file(p).plot_(**kws)
        return ax
    else:
        from roux.viz.image import plot_image
        if p.endswith('.py'):
            p=(read_ps(f"{dirname(p)}.*png")+read_ps(f"{dirname(p)}.*pdf"))[0]
        return plot_image(p,**kws)

## files
def to_concat(
    ps: list,
    how: str='h',
    use_imagemagick: bool= False,
    use_conda_env: bool= False,
    test: bool=False,
    **kws_outp,
    ) -> str:
    """Concat images.

    Args:
        ps (list): list of paths.
        how (str, optional): horizontal (`h`) or vertical `v`. Defaults to 'h'.
        test (bool, optional): test mode. Defaults to False.

    Returns:
        str: path of the output.
    """
    outp=to_outp(ps,**kws_outp)
    if use_imagemagick:
        com=f"convert {'+' if how=='h' else '-'}append {' '.join(ps)} {outp}"
        if use_conda_env:
            env=get_env('imagemagick')
            runbash(
                f"{env['PATH'].split(':')[0]}/{com}",
                env=env,
                test=test,
               )
        else:
            if test:
                logging.info(com)
            response=subprocess.call(com,shell=True)
            assert response==0 and exists(outp)
    else:
        import sys
        from PIL import Image
        images = [Image.open(x) for x in ps]
    #     widths, heights = zip(*(i.size for i in images))
    #     total_width = sum(widths)
    #     max_height = max(heights)
        if how=='h':
            imgs_comb = Image.fromarray( images)
            # for a vertical stacking it is simple: use vstack
        elif how=='v':
            min_shape = sorted( [(np.sum(i.size), i.size ) for i in images])[0][1]
            imgs_comb = Image.fromarray( np.vstack( (np.asarray( i.resize(min_shape) ) for i in images ) ))
        else:
            raise ValueError(how)
        imgs_comb.save(outp) 
    return outp

def to_montage(
    ps: list,
    layout: list,
    hspace: float=0,
    vspace: float=0,
    test: bool=False,
    **kws_outp
    ) -> str:
    """To montage.

    Args:
        ps (_type_): list of paths.
        layout (_type_): layout of the images.
        hspace (int, optional): horizontal space. Defaults to 0.
        vspace (int, optional): vertical space. Defaults to 0.
        test (bool, optional): test mode. Defaults to False.

    Returns:
        str: path of the output.
    """
    outp=to_outp(ps,**kws_outp)
    env=get_env('imagemagick')
    runbash(
        f"{env['PATH'].split(':')[0]}/montage -geometry +{hspace}+{vspace} -tile {layout} {' '.join(ps)} {outp}",
        env=env,
        test=test,
       )
    return outp

def to_gif(
    ps: list,
    outp: str,
    duration: int=200,
    loop: int=0,
    optimize: bool=True
    ) -> str:
    """Convert to GIF.

    Args:
        ps (list): list of paths.
        outp (str): output path.
        duration (int, optional): duration. Defaults to 200.
        loop (int, optional): loop or not. Defaults to 0.
        optimize (bool, optional): optimize the size. Defaults to True.

    Returns:
        str: output path.

    References:
        1. https://pillow.readthedocs.io/en/stable/handbook/image-file-formats.html#gif
        2. https://stackoverflow.com/a/57751793/3521099
    """
    import glob
    from PIL import Image
    img, *imgs = [Image.open(f) for f in sorted(glob.glob(ps) if isinstance(ps,str) else ps)]
    makedirs(outp)
    width, height = imgs[0].size
    img=img.resize((width//5, height//5))
    imgs=[im.resize((width//5, height//5)) for im in imgs]
    img.save(fp=outp, format='GIF', append_images=imgs,
             save_all=True,
             duration=duration, loop=loop,
          optimize=optimize )    
    
def to_data(path: str) -> str:
    """Convert to base64 string.

    Args:
        path (str): path of the input.

    Returns:
        base64 string.
    """
    import base64
    with open(path, "rb") as image_file:
        encoded_string = "data:image/jpeg;base64,"+base64.b64encode(image_file.read()).decode()
    return encoded_string

from os.path import exists,basename,dirname
# from roux.lib.sys import runbash
import subprocess
from roux.lib.io import makedirs
from glob import glob
import logging

def to_convert(filep: str,outd: str=None,fmt: str="JPEG") -> str:
    """Convert format of image using `PIL`.

    Args:
        filep (str): input path.
        outd (str, optional): output directory. Defaults to None.
        fmt (str, optional): format of the output. Defaults to "JPEG".

    Returns:
        str: output path.
    """
    from PIL import Image
    im = Image.open(filep)
    if not outd is None:
        outp=f"{outd}/{basename(filep)}.{fmt.lower()}"
    else:
        outp=filep+".{fmt.lower()}"
    im.convert('RGB').save(outp,fmt)
    return outp
    
def to_raster(
    plotp: str,
    dpi: int=500,
    alpha: bool=False,
    trim: bool=False,
    force: bool=False,
    test: bool=False
    ) -> str:
    """to_raster _summary_

    Args:
        plotp (str): input path.
        dpi (int, optional): DPI. Defaults to 500.
        alpha (bool, optional): transparency. Defaults to False.
        trim (bool, optional): trim margins. Defaults to False.
        force (bool, optional): overwrite output. Defaults to False.
        test (bool, optional): test mode. Defaults to False.

    Returns:
        str: _description_
    
    Notes: 
        1. Runs a bash command: `convert -density 300 -trim`.
    """
    plotoutp=f"{plotp}.png"
    if not exists(plotp):
        logging.error(f'{plotp} not found')
        return
    if not exists(plotoutp) or force: 
        import subprocess
        try:
            toolp = subprocess.check_output(["which", "convert"]).strip()
        except subprocess.CalledProcessError:
            logging.error('make sure imagemagick is installed. conda install imagemagick')
            return
        com=f'convert -density 500 '+('-background none ' if alpha else '')+'-interpolate Catrom -resize "2000" '+('-trim ' if trim else '')+f"{plotp} {plotoutp}"
        subprocess.call(com,shell=True)
    return plotoutp
    
def to_rasters(plotd,ext='svg'):
    """Convert many images to raster. Uses inkscape.

    Args:
        plotd (str): directory.
        ext (str, optional): extension of the output. Defaults to 'svg'.
    """
    logging.info(glob(f"{plotd}/*.{ext}"))
#     plotd='plot/'
#     com=f'for f in {plotd}/*.{ext}; do convert -density 500 -alpha off -resize "2000" -trim "$f" "$f.png"; done'
    com=f'for f in {plotd}/*.{ext}; do inkscape "$f" -z --export-dpi=500 --export-area-drawing --export-png="$f.png"; done'
    return subprocess.call(com,shell=True)


import pandas as pd
import numpy as np
from scipy import stats

def norm_by_quantile(X: np.array) -> np.array:
    """Normalize the columns of X to each have the same distribution.

    Notes:
        Given an expression matrix (microarray data, read counts, etc) of M genes
        by N samples, quantile normalization ensures all samples have the same
        spread of data (by construction).

        The data across each row are averaged to obtain an average column. Each
        column quantile is replaced with the corresponding quantile of the average
        column.

    Parameters:
        X : 2D array of float, shape (M, N). The input data, with M rows (genes/features) and N columns (samples).

    Returns:
        Xn : 2D array of float, shape (M, N). The normalized data.
    """
    # compute the quantiles
    quantiles = np.mean(np.sort(X, axis=0), axis=1)

    # compute the column-wise ranks. Each observation is replaced with its
    # rank in that column: the smallest observation is replaced by 1, the
    # second-smallest by 2, ..., and the largest by M, the number of rows.
    ranks = np.apply_along_axis(stats.rankdata, 0, X)

    # convert ranks to integer indices from 0 to M-1
    rank_indices = ranks.astype(int) - 1

    # index the quantiles for each rank with the ranks matrix
    Xn = quantiles[rank_indices]
    
    if isinstance(X,pd.DataFrame):
        return pd.DataFrame(Xn,columns=X.columns,index=X.index)
    else:
        return(Xn)

def norm_by_gaussian_kde(values: np.array) -> np.array:
    """Normalise matrix by gaussian KDE.

    Args:
        values (np.array): input matrix.

    Returns:
        np.array: output matrix.

    References: 
        https://github.com/saezlab/protein_attenuation/blob/6c1e81af37d72ef09835ee287f63b000c7c6663c/src/protein_attenuation/utils.py
    """
    kernel = stats.gaussian_kde(values)
    return Series({k: np.log(kernel.integrate_box_1d(-1e4, v) / kernel.integrate_box_1d(v, 1e4)) for k, v in values.to_dict().items()})


## z-scores

def zscore(df,cols: list=None):
    """Z-score.

    Args:
        ds (pd.Series): input vector.

    Returns:
        pd.Series: output vector.
    """
    if isinstance(df, pd.Series):
        return (df - df.mean())/df.std()
    else:
        if cols is None:
            cols=df.columns
        return (df[cols] - df[cols].mean())/df[cols].std()
    

def zscore_robust(a: np.array) -> np.array:
    """Robust Z-score.

    Args:
        a (np.array): input data.

    Returns:
        np.array: output.
        
    Example:
        t = sc.stats.norm.rvs(size=100, scale=1, random_state=123456)
        plt.hist(t,bins=40)
        plt.hist(apply_zscore_robust(t),bins=40)
        print(np.median(t),np.median(apply_zscore_robust(t)))
    """
    def get_zscore_robust(x,median,mad):
        return (x-median)/(mad*1.4826)
    median=np.median(a)
    mad=sc.stats.median_abs_deviation(a)
    if mad==0:
        logging.error('mad==0')
    return [get_zscore_robust(x,median,mad) for x in a]







# maintain back-compatibility using aliases
# import sys
# import logging

# d1={'roux.lib.cloud.google': 'roux.lib.google', 'roux.lib.code.df': 'roux.workflow.df', 'roux.lib.code.function': 'roux.workflow.function', 'roux.lib.code.io': 'roux.workflow.io', 'roux.lib.code.knit': 'roux.workflow.knit', 
#     # 'roux.lib.code.monitor': 'roux.workflow.monitor', 
#     'roux.lib.code.version': 'roux.workflow.version', 'roux.lib.code.workflow': 'roux.workflow.workflow', 
#     # 'roux.lib.database.biomart': 'roux.query.biomart',
#     # 'roux.lib.database.ensembl': 'roux.query.ensembl',
#     # 'roux.lib.figure.figure': 'roux.viz.figure',
#     # 'roux.lib.figure.io': 'roux.viz.io',
#     'roux.lib.io_df': 'roux.lib.df', 'roux.lib.io_dfs': 'roux.lib.dfs', 'roux.lib.io_dict': 'roux.lib.dict', 'roux.lib.io_files': 'roux.lib.io', 
#     # 'roux.lib.io_seqs': 'roux.lib.seq',
#     'roux.lib.io_sets': 'roux.lib.set', 'roux.lib.io_strs': 'roux.lib.str', 'roux.lib.io_sys': 'roux.lib.sys', 'roux.lib.io_text': 'roux.lib.text',
#     # 'roux.lib.plot.annot': 'roux.viz.annot', 'roux.lib.plot.ax_': 'roux.viz.ax_', 'roux.lib.plot.bar': 'roux.viz.bar', 'roux.lib.plot.colors': 'roux.viz.colors', 'roux.lib.plot.dist': 'roux.viz.dist', 'roux.lib.plot.heatmap': 'roux.viz.heatmap', 'roux.lib.plot.image': 'roux.viz.image', 'roux.lib.plot.line': 'roux.viz.line', 'roux.lib.plot.scatter': 'roux.viz.scatter', 'roux.lib.plot.sequence': 'roux.viz.sequence', 'roux.lib.plot.sets': 'roux.viz.sets', 'roux.lib.stat.binary': 'roux.stat.binary', 'roux.lib.stat.classify': 'roux.stat.classify', 'roux.lib.stat.cluster': 'roux.stat.cluster', 'roux.lib.stat.corr': 'roux.stat.corr', 'roux.lib.stat.diff': 'roux.stat.diff', 'roux.lib.stat.enrich': 'roux.stat.enrich', 'roux.lib.stat.fit': 'roux.stat.fit', 'roux.lib.stat.io': 'roux.stat.io', 'roux.lib.stat.network': 'roux.stat.network', 'roux.lib.stat.norm': 'roux.stat.norm', 'roux.lib.stat.paired': 'roux.stat.paired', 'roux.lib.stat.solve': 'roux.stat.solve', 'roux.lib.stat.transform': 'roux.stat.transform', 'roux.lib.stat.variance': 'roux.stat.variance',
#    }

# for k,v in d1.items():
#     try:
#         import importlib
#         importlib.import_module(v)
#         sys.modules[k] = sys.modules[v]
#     except:
#         logging.warning(f"import path '{k}' is not available.")


from roux.global_imports import *

# stats
# def annot_stats(dplot: pd.DataFrame,
#             colx: str,
#             coly: str,
#             colz: str,
#             stat_method: list=[],
#             bootstrapped: bool=False,
#             params_set_label: dict={},
#             ax: plt.Axes = None,
#             ) -> plt.Axes:
#     """Annotate stats on a scatter plot.

#     Args:
#         dplot (pd.DataFrame): input dataframe.
#         colx (str): x column.
#         coly (str): y column.
#         colz (str): z column.
#         stat_method (list, optional): names of stat methods to apply. Defaults to [].
#         bootstrapped (bool, optional): to bootstrap the data or not. Defaults to False.
#         params_set_label (dict, optional): parameters provided to the `set_label` function. Defaults to {}.
#         ax (plt.Axes, optional): `plt.Axes` object. Defaults to None.

#     Returns:
#         plt.Axes: `plt.Axes` object.
#     """
#     stat_method = [stat_method] if isinstance(stat_method,str) else stat_method
#     from roux.viz.annot import set_label
#     if 'mlr' in stat_method:
#         from roux.lib.stat.poly import get_mlr_2_str
#         ax=set_label(ax,label=get_mlr_2_str(dplot,colz,[colx,coly]),
#                     title=True,params={'loc':'left'})
#     if 'spearman' in stat_method or 'pearson' in stat_method:
#         from roux.stat.corr import get_corr
#         ax=set_label(ax,label=get_corr(dplot[colx],dplot[coly],method=stat_method[0],
#                                        bootstrapped=bootstrapped,
#                                        outstr=True,n=True),
#                     **params_set_label)
#     return ax

def plot_trendline(dplot: pd.DataFrame,
                    colx: str,
                    coly: str,
                    params_plot: dict={'color':'r','lw':2},
                    poly: bool=False,
                    lowess: bool=True,
                    linestyle: str= 'solid',
                    params_poly: dict={'deg':1},
                    params_lowess: dict={'frac':0.7,'it':5},
                    ax: plt.Axes = None,
                  **kws
            ) -> plt.Axes:
    """Plot a trendline.

    Args:
        dplot (pd.DataFrame): input dataframe.
        colx (str): x column.
        coly (str): y column.
        params_plot (dict, optional): parameters provided to the plot. Defaults to {'color':'r','linestyle':'solid','lw':2}.
        poly (bool, optional): apply polynomial function. Defaults to False.
        lowess (bool, optional): apply lowess function. Defaults to True.
        params_poly (_type_, optional): parameters provided to the polynomial function. Defaults to {'deg':1}.
        params_lowess (_type_, optional): parameters provided to the lowess function.. Defaults to {'frac':0.7,'it':5}.
        ax (plt.Axes, optional): `plt.Axes` object. Defaults to None.

    Keyword Args:
        kws: parameters provided to the `plot` function. 

    Returns:
        plt.Axes: `plt.Axes` object.

    TODOs: 
        1. Label with goodness of fit, r (y_hat vs y)
    """
    ax= plt.subplot() if ax is None else ax    
    if poly:
        coef = np.polyfit(dplot[colx], dplot[coly],**params_poly)
        poly1d_fn = np.poly1d(coef)
        # poly1d_fn is now a function which takes in x and returns an estimate for y
        ax.plot(dplot[colx], poly1d_fn(dplot[colx]),linestyle=linestyle, **params_plot,**kws)
    if lowess:
        from statsmodels.nonparametric.smoothers_lowess import lowess
        xys_lowess=lowess(dplot[coly], dplot[colx],frac=0.7,it=5)
        ax.plot(xys_lowess[:,0],xys_lowess[:,1],linestyle=linestyle, **params_plot,**kws)
    return ax

# @to_class(rd)
def plot_scatter(
    dplot: pd.DataFrame,
    colx: str,
    coly: str,
    colz: str=None,
    kind: str='scatter',
    trendline_method: str='poly',
    stat_method: str="spearman",
    bootstrapped: bool=False,
    cmap: str='Reds',
    label_colorbar: str=None,
    gridsize: int=25,
    bbox_to_anchor: list=[1,1],
    loc: str='upper left',
    title: str=None,
    params_plot: dict={},
    params_plot_trendline: dict={},
    params_set_label: dict={},
    ax: plt.Axes = None,
    **kws,
    ) -> plt.Axes:
    """Plot scatter.

    Args:
        dplot (pd.DataFrame): input dataframe.
        colx (str): x column.
        coly (str): y column.
        colz (str, optional): z column. Defaults to None.
        kind (str, optional): kind of scatter. Defaults to 'hexbin'.
        trendline_method (str, optional): trendline method ['poly','lowess']. Defaults to 'poly'.
        stat_method (str, optional): method of annoted stats ['mlr',"spearman"]. Defaults to "spearman".
        bootstrapped (bool, optional): bootstrap data. Defaults to False.
        cmap (str, optional): colormap. Defaults to 'Reds'.
        label_colorbar (str, optional): label of the colorbar. Defaults to None.
        gridsize (int, optional): number of grids in the hexbin. Defaults to 25.
        bbox_to_anchor (list, optional): location of the legend. Defaults to [1,1].
        loc (str, optional): location of the legend. Defaults to 'upper left'.
        title (str, optional): title of the plot. Defaults to None.
        params_plot (dict, optional): parameters provided to the `plot` function. Defaults to {}.
        params_plot_trendline (dict, optional): parameters provided to the `plot_trendline` function. Defaults to {}.
        params_set_label (dict, optional): parameters provided to the `set_label` function. Defaults to dict(x=0,y=1).
        ax (plt.Axes, optional): `plt.Axes` object. Defaults to None.

    Keyword Args:
        kws: parameters provided to the `plot` function. 

    Returns:
        plt.Axes: `plt.Axes` object.

    Notes:
        For a rasterized scatter plot set `scatter_kws={'rasterized': True}`
    
    TODOs:
        1. Access the function as an attribute of roux-data i.e. `rd`. 

    """
    ax= plt.subplot() if ax is None else ax
    
    trendline_method = [trendline_method] if isinstance(trendline_method,str) else [] if trendline_method is None else trendline_method
    stat_method = [stat_method] if isinstance(stat_method,str) else [] if stat_method is None else stat_method
    
    dplot=dplot.dropna(subset=[colx,coly]+[] if colz is None else [colz],how='any')
    if kind in ['hexbin']:
        if colz is None: 
            colz='count'
            dplot[colz]=1
        if colz in dplot:
            params_plot['C']=colz
            params_plot['reduce_C_function']=len if colz=='count' else params_plot['reduce_C_function'] if 'reduce_C_function' in params_plot else np.mean        
            params_plot['gridsize']=params_plot['gridsize'] if 'gridsize' in params_plot else gridsize
            params_plot['cmap']=params_plot['cmap'] if 'cmap' in params_plot else cmap
            print(params_plot)
        ax=dplot.plot(kind=kind, x=colx, y=coly, 
    #         C=colz,
            ax=ax,
            **params_plot,
                      **kws,
            )
    else:
        ax=sns.scatterplot(data=dplot,
                            x=colx,
                            y=coly,
                            hue=colz,       
                            palette=cmap,
                            ax=ax,
                           **params_plot,
                           **kws)
        if not colz is None:
            leg=ax.legend(loc=loc,bbox_to_anchor=bbox_to_anchor,title=colz if title is None else title)
            if '\n' in title:
                leg._legend_box.align = "center"
    from roux.viz.ax_ import set_colorbar_label
#     print(colz)
    ax=set_colorbar_label(ax,colz if label_colorbar is None else label_colorbar)
    from roux.viz.annot import set_label
    if 'mlr' in stat_method:
        from roux.lib.stat.poly import get_mlr_2_str
        ax=set_label(ax,label=get_mlr_2_str(dplot,colz,[colx,coly]),
                    title=True,params={'loc':'left'})
    if 'spearman' in stat_method or 'pearson' in stat_method:
        from roux.stat.corr import get_corr
        label,r=get_corr(dplot[colx],dplot[coly],method=stat_method[0],
                                       bootstrapped=bootstrapped,
                                       outstr=True,
                                      # n=True
                                     )
        if not 'loc' in params_set_label:
            if r>=0:
                params_set_label['loc']=2
            elif r<0:
                params_set_label['loc']=1
        ax=set_label(ax=ax,s=label,**params_set_label)
    from roux.viz.colors import saturate_color
    plot_trendline(dplot,colx,coly,
                    params_plot={'color':saturate_color(params_plot['color'],alpha=1.75) if 'color' in params_plot else None,},
                    poly='poly' in trendline_method,
                    lowess='lowess' in trendline_method,
                   ax=ax, 
                   **params_plot_trendline,
                    )    
    return ax
    
def plot_qq(
    x: pd.Series
    ) -> plt.Axes:
    """plot QQ.

    Args:
        x (pd.Series): input vector.

    Returns:
        plt.Axes: `plt.Axes` object.
    """
    import statsmodels.api as sm
    fig = plt.figure(figsize = [3, 3])
    ax = plt.subplot()
    sm.qqplot(x, dist = sc.stats.norm, 
              line = 's', 
              ax=ax)
    ax.set_title("SW test "+pval2annot(sc.stats.shapiro(x)[1],alpha=0.05,fmt='<',linebreak=False))
    from roux.viz.ax_ import set_equallim
    ax=set_equallim(ax)
    return ax

def plot_ranks(
    df1: pd.DataFrame,
    colid: str,
    colx: str,
    coly: str='rank',
    ascending: bool=True,
    # line: bool=False,
    ax=None,
    **kws,
    ) -> plt.Axes:
    """Plot rankings.

    Args:
        dplot (pd.DataFrame): input data.
        colx (str): x column.
        coly (str): y column.
        colid (str): column with unique ids.
        ax (plt.Axes, optional): `plt.Axes` object. Defaults to None.

    Keyword Args:
        kws: parameters provided to the `seaborn.scatterplot` function. 

    Returns:
        plt.Axes: `plt.Axes` object.
    """    
    assert not df1[colid].duplicated().any()
    df1[coly]=df1[colx].rank(ascending=ascending)
    if ax is None:
        fig,ax=plt.subplots(figsize=[2,2])
    ax=sns.scatterplot(data=df1,
                   x=colx,y=coly,
                       **kws,
                   ax=ax)
    # if line:
    ax.set(
        yticks=[int(i) for i in np.linspace(1,len(df1),4)], ## start with 1
            )
    if ascending:
        ax.invert_yaxis()
    return ax


from roux.global_imports import *
from roux.viz.colors import *
from roux.viz.annot import *
from roux.viz.ax_ import *

## single distributions.
def hist_annot(
    dplot: pd.DataFrame,
    colx: str,
    colssubsets: list=[],
    bins: int=100,
    subset_unclassified: bool=True,
    cmap: str='hsv',
    ymin=None,
    ymax=None,
    ylimoff: float=1,
    ywithinoff: float=1.2,
    annotaslegend: bool=True,
    annotn: bool=True,
    params_scatter: dict={'zorder':2,'alpha':0.1,'marker':'|'},
    xlim: tuple=None,
    ax: plt.Axes = None,
    **kws,
    ) -> plt.Axes:
    """Annoted histogram.

    Args:
        dplot (pd.DataFrame): input dataframe.
        colx (str): x column.
        colssubsets (list, optional): columns indicating subsets. Defaults to [].
        bins (int, optional): bins. Defaults to 100.
        subset_unclassified (bool, optional): call non-annotated subset as 'unclassified'. Defaults to True.
        cmap (str, optional): colormap. Defaults to 'Reds_r'.
        ylimoff (float, optional): y-offset for y-axis limit . Defaults to 1.2.
        ywithinoff (float, optional): y-offset for the distance within labels. Defaults to 1.2.
        annotaslegend (bool, optional): convert labels to legends. Defaults to True.
        annotn (bool, optional): annotate sample sizes. Defaults to True.
        params_scatter (_type_, optional): parameters of the scatter plot. Defaults to {'zorder':2,'alpha':0.1,'marker':'|'}.
        xlim (tuple, optional): x-axis limits. Defaults to None.
        ax (plt.Axes, optional): `plt.Axes` object. Defaults to None.

    Keyword Args:
        kws: parameters provided to the `hist` function.

    Returns:
        plt.Axes: `plt.Axes` object.
    
    TODOs:
        For scatter, use `annot_side` with `loc='top'`.
    """
    from roux.viz.ax_ import reset_legend_colors
    if not xlim is None:
        logging.warning('colx adjusted to xlim')
        dplot.loc[(dplot[colx]<xlim[0]),colx]=xlim[0]
        dplot.loc[(dplot[colx]>xlim[1]),colx]=xlim[1]
    if ax is None:ax=plt.subplot(111)
    ax=dplot[colx].hist(bins=bins,ax=ax,zorder=1,**kws,)
    ax.set_xlabel(colx)
    ax.set_ylabel('count')
    if not xlim is None:
        ax.set_xlim(xlim)
    ax.set_ylim(0,ax.get_ylim()[1]*ylimoff)        
    from roux.viz.colors import get_ncolors
    colors=get_ncolors(len(colssubsets),cmap=cmap)
    for colsubsetsi,(colsubsets,color) in enumerate(zip(colssubsets,colors)):
        subsets=[s for s in dropna(dplot[colsubsets].unique()) if not (subset_unclassified and s=='unclassified')]
        for subseti,subset in enumerate(subsets):
            y=((ax.set_ylim()[1] if ymax is None else ymax) - (ax.set_ylim()[0] if ymin is None else ymin))*((10-(subseti*ywithinoff+colsubsetsi))/10-0.05)+ax.set_ylim()[0]
            X=dplot.loc[(dplot[colsubsets]==subset),colx]
            Y=[y for i in X]
            ax.scatter(X,Y,
                       color=color,**params_scatter)
            ax.text((max(X) if not annotaslegend else ax.get_xlim()[1]),
                    max(Y),
                    f"{subset}\n(n={len(X)})" if annotn else f" {subset}",
                    ha='left',va='center')
    #     break
#     ax=reset_legend_colors(ax)
#     ax.legend(bbox_to_anchor=[1,1])
    return ax
        
def plot_gmm(
    x: pd.Series,
    coff: float=None,
    mix_pdf: object=None,
    two_pdfs: tuple=None,
    weights: tuple=None,
    n_clusters: int=2,
    bins: int=20,
    show_cutoff: bool=True,
    show_cutoff_line: bool=True,
    colors: list=['gray','gray','lightgray'],
    out_coff: bool= False,
    hist: bool=True,
    test: bool=False,
    ax: plt.Axes = None,
    kws_axvline=dict(color='k'),
    **kws,
    ) -> plt.Axes:
    """Plot Gaussian mixture Models (GMMs).

    Args:
        x (pd.Series): input vector.
        coff (float, optional): intersection between two fitted distributions. Defaults to None.
        mix_pdf (object, optional): Probability density function of the mixed distribution. Defaults to None.
        two_pdfs (tuple, optional): Probability density functions of the separate distributions. Defaults to None.
        weights (tuple, optional): weights of the individual distributions. Defaults to None.
        n_clusters (int, optional): number of distributions. Defaults to 2.
        bins (int, optional): bins. Defaults to 50.
        colors (list, optional): colors of the invividual distributions and of the mixed one. Defaults to ['gray','gray','lightgray'].
        'gray'
        out_coff (bool,False): return the cutoff. Defaults to False.
        hist (bool, optional): show histogram. Defaults to True.
        test (bool, optional): test mode. Defaults to False.
        ax (plt.Axes, optional): `plt.Axes` object. Defaults to None.
        
    Keyword Args:
        kws: parameters provided to the `hist` function.
        kws_axvline: parameters provided to the `axvline` function.

    Returns:
        plt.Axes: `plt.Axes` object.
    """
    if mix_pdf is None:
        from roux.stat.cluster import cluster_1d
        d_=cluster_1d(x,
                n_clusters=n_clusters,
                clf_type='gmm',
                random_state=88,
                test=False,
                returns=['coff','mix_pdf','two_pdfs','weights'],
                )
        coff,mix_pdf,two_pdfs,weights=d_['coff'],d_['mix_pdf'],d_['two_pdfs'],d_['weights']
    if ax is None:
        plt.figure(figsize=[2.5,2.5])
        ax=plt.subplot()
    # plot histogram
    if hist:
        pd.Series(x).hist(density=True,
                          histtype='step',
                          bins=bins,
                          ax=ax,
                          **kws)
    # plot fitted distributions
    ax.plot(x,mix_pdf.ravel(), c=colors[-1])
    for i in range(n_clusters):
        ax.plot(x,two_pdfs[i]*weights[i], c=colors[i])
#     ax.plot(x,two_pdfs[1]*weights[1], c='gray')
    if n_clusters==2:
        if show_cutoff_line:
            ax.axvline(coff,**kws_axvline)
        if show_cutoff:
            ax.text(coff,ax.get_ylim()[1],f"{coff:.1f}",ha='center',va='bottom')
    if not out_coff:
        return ax
    else:
        return ax,coff

def plot_normal(
    x: pd.Series,
    ax: plt.Axes = None,
    ) -> plt.Axes:
    """Plot normal distribution.

    Args:
        x (pd.Series): input vector.
        ax (plt.Axes, optional): `plt.Axes` object. Defaults to None.

    Returns:
        plt.Axes: `plt.Axes` object.
    """
    if not ax is None:
        fig,ax = plt.subplots(figsize = [3, 3])
    import statsmodels.api as sm
    ax = sns.distplot(x, hist = True, 
                      kde_kws = {"shade" : True, "lw": 1, }, 
                      fit = sc.stats.norm,
                      label='residuals',
                     )
    ax.set_title("SW test "+pval2annot(sc.stats.shapiro(x)[1],alpha=0.05,fmt='<',linebreak=False))
    ax.legend()
    return ax

## paired distributions.
def plot_dists(
    df1: pd.DataFrame,
    x: str,
    y: str,
    colindex: str,
    hue: str=None,
    order: list=None,
    hue_order: list=None,
    kind: str='box',
    show_p: bool=True,
    show_n: bool=True,
    show_n_prefix: str='',
    show_n_ha='right',
    alternative: str='two-sided',
    offx_n: float=0,
    xlim: tuple=None,
    offx_pval: float=0.05,
    offy_pval: float=None,
    saturate_color_alpha: float=1.5,
    ax: plt.Axes = None,
    test: bool= False,
    kws_stats: dict={},
    **kws
    ) -> plt.Axes:
    """Plot distributions.

    Args:
        df1 (pd.DataFrame): input data.
        x (str): x column.
        y (str): y column.
        colindex (str): index column.
        hue (str, optional): column with values to be encoded as hues. Defaults to None.
        order (list, optional): order of categorical values. Defaults to None.
        hue_order (list, optional): order of values to be encoded as hues. Defaults to None.
        kind (str, optional): kind of distribution. Defaults to 'box'.
        show_p (bool, optional): show p-values. Defaults to True.
        show_n (bool, optional): show sample sizes. Defaults to True.
        show_n_prefix (str, optional): show prefix of sample size label i.e. `n=`. Defaults to ''.
        offx_n (float, optional): x-offset for the sample size label. Defaults to 0.
        xlim (tuple, optional): x-axis limits. Defaults to None.
        offx_pval (float, optional): x-offset for the p-value labels. Defaults to 0.05.
        offy_pval (float, optional): y-offset for the p-value labels. Defaults to None.
        saturate_color_alpha (float, optional): saturation of the color. Defaults to 1.5.
        ax (plt.Axes, optional): `plt.Axes` object. Defaults to None.
        test (bool, optional): test mode. Defaults to False.
        kws_stats (dict, optional): parameters provided to the stat function. Defaults to {}.

    Keyword Args:
        kws: parameters provided to the `seaborn` function. 

    Returns:
        plt.Axes: `plt.Axes` object.
        
    TODOs:
        1. Sort categories.
        2. Change alpha of the boxplot rather than changing saturation of the swarmplot. 

    """
    if isinstance(colindex,str):
        colindex=[colindex]
    df1=df1.log.dropna(subset=colindex+[x,y])
    df1[y]=df1[y].astype(str)
    if order is None:
        order=df1[y].unique().tolist()
    if not hue is None and hue_order is None:
        hue_order=df1[hue].unique().tolist()
    if (hue is None) and (isinstance(show_p,bool) and show_p):
        from roux.stat.diff import get_stats
        df2=get_stats(df1,
                      colindex=colindex,
                          colsubset=y,
                          cols_value=[x],
                          subsets=order,
        #                   alpha=0.05
                        axis=0,
                        **kws_stats,
                         )
        if df2 is None:
            logging.error("get_stats failed.")
            d1={}
        else:
            df2=df2.reset_index()
            # df1=df1.rd.renameby_replace({f"{} ":''})
            df2=df2.loc[(df2['subset1']==order[0]),:]
            # print(df2)
            d1=df2.rd.to_dict(['subset2','P (MWU test)'])
    elif (not hue is None) and (isinstance(show_p,bool) and show_p):
        from roux.stat.diff import get_stats_groupby
        df2=get_stats_groupby(df1.loc[df1[hue].isin(hue_order),:],
                          cols=[y],
                          colsubset=hue,
                          cols_value=[x],
                          colindex=colindex,
                          alpha=0.05,
                          axis=0,
                          **kws_stats,
                         ).reset_index()
        # df1=df1.rd.renameby_replace({f"{} ":''})
        if test:
            print(df2['subset1'])
        # df2=df2.loc[(df2['subset1']==hue_order[1]),:]
        # d1=df2.rd.to_dict([y,'P (MWU test)'])
        d1=df2.set_index(y)['P (MWU test)'].to_dict()
        if test:
            print(d1)
    if ax is None:
        _,ax=plt.subplots(figsize=[2,2])
    if isinstance(kind,str):
        kind={kind:{}}
    elif isinstance(kind,list):
        kind={k:{} for k in kind}
    for k in kind:
        kws_=kws.copy()
        # print(kws['palette'],kind)
        # if 'palette' in kws and any([k_ in kind for k_ in ['swarm','strip']]):
        #     from roux.viz.colors import saturate_color
        #     kws['palette']=[saturate_color(color=c, alpha=saturate_color_alpha-1) for c in kws['palette']]            
        # if 'palette' in kws and k in ['swarm','strip']:
            # from roux.viz.colors import saturate_color
            # kws['palette']=[saturate_color(color=c, alpha=saturate_color_alpha+0.5) for c in kws['palette']]
        if k=='box' and (('swarm' in kind) or ('strip' in kind)):
            kws_['boxprops']=dict(alpha=.3)
            # print(kws_,kws)
            # print(kws['palette'])
        getattr(sns,k+"plot")(data=df1,
                    x=x,y=y,
                    hue=hue,
                    order=order,
                    hue_order=hue_order,
                    **kind[k],
                    **kws_,
                     ax=ax)
    ax.set(xlabel=x)
    d2=get_ticklabel2position(ax,'y')
    ax.set(
          ylabel=None if hue is None else y,
          xlim=xlim,
          )
    d3=get_axlims(ax)
    if isinstance(show_p,(bool,dict)):
        if isinstance(show_p,bool) and show_p:
            d1={k:pval2annot(d1[k],alternative=alternative,fmt='<',linebreak=False) for k in d1}
        else:
            d1=show_p
        if offy_pval is None and hue is None:
            offy_pval=-0.5
        else:
            offy_pval=0            
        if isinstance(d1,dict):
            for k,s in d1.items():
                if test:
                    print(d1)
                    print(d2)
                    print(d3)
                ax.text(d3['x']['max']+(d3['x']['len']*offx_pval),d2[k]+offy_pval,s,va='center')
    if show_n:
        df1_=df1.groupby(y).apply(lambda df: df.groupby(colindex).ngroups).to_frame('n').reset_index()
        df1_['y']=df1_[y].map(d2)
        import matplotlib.transforms as transforms
        df1_.apply(lambda x: ax.text(x=1.1+offx_n,y=x['y'],
                                   s=show_n_prefix+str(x['n']),va='center',ha=show_n_ha,
                                     transform=transforms.blended_transform_factory(ax.transAxes,ax.transData),
                                   ),axis=1)
    ax.tick_params(axis='y', colors='k')
    if not hue is None:
        o1=ax.legend(
                  loc='upper left', 
                  bbox_to_anchor=(1, 0),
            frameon=True,
            title=hue,
            )
        o1.get_frame().set_edgecolor((0.95,0.95,0.95))
    return ax

def pointplot_groupbyedgecolor(
    data: pd.DataFrame,
    ax: plt.Axes = None,
    **kws
    ) -> plt.Axes:
    """Plot seaborn's `pointplot` grouped by edgecolor of points.

    Args:
        data (pd.DataFrame): input data.
        ax (plt.Axes, optional): `plt.Axes` object. Defaults to None.
    
    Keyword Args:
        kws: parameters provided to the `seaborn`'s `pointplot` function. 
        
    Returns:
        plt.Axes: `plt.Axes` object.
    """
    ax=plt.subplot() if ax is None else ax
    ax=sns.pointplot(data=data,
                     ax=ax,
                     **kws)
    plt.setp(ax.collections, sizes=[100])   
    for c in ax.collections:
        if c.get_label().startswith(kws['hue_order'][0].split(' ')[0]):
            c.set_linewidth(2)
            c.set_edgecolor('k')
        else:
            c.set_linewidth(2)        
            c.set_edgecolor('w')
    ax.legend(bbox_to_anchor=[1,1])
    return ax    


from roux.global_imports import *
from roux.viz.ax_ import *
from roux.viz.annot import *

def plot_venn(
    ds1: pd.Series,
    ax: plt.Axes=None,
    figsize: tuple=[2.5,2.5],
    show_n: bool=True
    **kws,
    ) -> plt.Axes:
    """Plot Venn diagram.

    Args:
        ds1 (pd.Series): input vector. Subsets in the index levels, mapped to counts. 
        ax (plt.Axes, optional): `plt.Axes` object. Defaults to None.
        figsize (tuple, optional): figure size. Defaults to [2.5,2.5].
        show_n (bool, optional): show sample sizes. Defaults to True.

    Returns:
        plt.Axes: `plt.Axes` object.
        
    Notes:
        1. Create the input pd.Series from dict.
        
            df_=to_map_binary(dict2df(d_).explode('value'),colgroupby='key',colvalue='value')
            ds_=df_.groupby(df_.columns.tolist()).size()
    """
    assert isinstance(ds1,pd.Series)
    assert ds1.dtypes=='int'
    assert len(ds1.index.names)>=2
    if ax is None:
        fig,ax=plt.subplots(figsize=figsize)
    if show_n:
        from roux.lib.df import get_totals
        d1=get_totals(ds1)
        ds1.index.names=[f"{k}\n({d1[k]})" for k in ds1.index.names]
    set_labels=list(ds1.index.names)
    if len(set_labels)==1 or len(set_labels)>3:
        logging.warning("need 2 or 3 sets")
        return 
    ds1.index=[''.join([str(int(i)) for i in list(t)]) for t in ds1.index]
    import matplotlib_venn as mv
    _=getattr(mv,f"venn{len(set_labels)}")(subsets=ds1.to_dict(),
                                            set_labels=set_labels,
                                            ax=ax,
                                            **kws,
         )
    return ax

def plot_intersections(
    ds1: pd.Series,
    item_name: str=None,
    figsize: tuple=[4,4],
    text_width: float=2,
    yorder: list=None,
    sort_by: str='cardinality',
    sort_categories_by: str=None,#'cardinality',
    element_size: int=40,
    facecolor: str='gray',
    bari_annot: int=None, # 0, 'max_intersections'
    totals_bar: bool=False,
    totals_text: bool=True,                           
    intersections_ylabel: float=None,
    intersections_min: float=None,
    test: bool=False,
    annot_text: bool=False,
    set_ylabelx: float=-0.25,
    set_ylabely: float=0.5,
    **kws,
    ) -> plt.Axes:
    """Plot upset plot.

    Args:
        ds1 (pd.Series): input vector.
        item_name (str, optional): name of items. Defaults to None.
        figsize (tuple, optional): figure size. Defaults to [4,4].
        text_width (float, optional): max. width of the text. Defaults to 2.
        yorder (list, optional): order of y elements. Defaults to None.
        sort_by (str, optional): sorting method. Defaults to 'cardinality'.
        sort_categories_by (str, optional): sorting method. Defaults to None.
        element_size (int, optional): size of elements. Defaults to 40.
        facecolor (str, optional): facecolor. Defaults to 'gray'.
        bari_annot (int, optional): annotate nth bar. Defaults to None.
        totals_text (bool, optional): show totals. Defaults to True.
        intersections_ylabel (float, optional): y-label of the intersections. Defaults to None.
        intersections_min (float, optional): intersection minimum to show. Defaults to None.
        test (bool, optional): test mode. Defaults to False.
        annot_text (bool, optional): annotate text. Defaults to False.
        set_ylabelx (float, optional): x position of the ylabel. Defaults to -0.25.
        set_ylabely (float, optional): y position of the ylabel. Defaults to 0.5.

    Keyword Args:
        kws: parameters provided to the `upset.plot` function. 

    Returns:
        plt.Axes: `plt.Axes` object.

    Notes:
        sort_by:{‘cardinality’, ‘degree’}
        If ‘cardinality’, subset are listed from largest to smallest. If ‘degree’, they are listed in order of the number of categories intersected.
        sort_categories_by:{‘cardinality’, None}
        Whether to sort the categories by total cardinality, or leave them in the provided order.

    References: 
        https://upsetplot.readthedocs.io/en/stable/api.html
    """
    assert(isinstance(ds1,pd.Series))
    if (item_name is None) and (not ds1.name is None):
        item_name=ds1.name
    if intersections_min is None:
        intersections_min=len(ds1)
    if not yorder is None:
        yorder= [c for c in yorder if c in ds1.index.names][::-1]
        ds1.index = ds1.index.reorder_levels(yorder) 
    ds2=(ds1/ds1.sum())*100
    import upsetplot as up
    d=up.plot(ds2,
              figsize=figsize,
              text_width=text_width,
              sort_by=sort_by,
              sort_categories_by=sort_categories_by,
              facecolor=facecolor,element_size=element_size,
              **kws,
             ) 
    d['totals'].set_visible(totals_bar)
    if totals_text:
        from roux.lib.df import get_totals
        d_=get_totals(ds1)
        d['matrix'].set_yticklabels([f"{s.get_text()} (n={d_[s.get_text()]})" for s in d['matrix'].get_yticklabels()],
                                          )
    if totals_bar:
        d['totals'].set(ylim=d['totals'].get_ylim()[::-1],
                       xlabel='%')
    set_ylabel(ax=d['intersections'],
              s=(f"{item_name}s " if not item_name is None else "")+f"%\n(total={ds1.sum()})",
               x=set_ylabelx, y=set_ylabely,
              )        
    d['intersections'].set(
                          xlim=[-0.5,intersections_min-0.5],
                          )
    if sort_by=='cardinality':
        y=ds2.max()
    elif sort_by=='degree':
        y=ds2.loc[tuple([True for i in ds2.index.names])]
#     if bari_annot=='max_intersections':
#         l1=[i for i,t in enumerate(ds1.index) if t==tuple(np.repeat(True,len(ds1.index.names)))]
#         if len(l1)==1:
#             bari_annot=l1[0]
#             print(bari_annot)
#     print(sum(ds1==ds1.max()))
#     print(bari_annot)
    if sum(ds1==ds1.max())!=1:
        bari_annot=None
    if isinstance(bari_annot,int):
        bari_annot=[bari_annot]
    if isinstance(bari_annot,list):
#         print(bari_annot)
        for i in bari_annot:
            d['intersections'].get_children()[i].set_color("#f55f5f")
    if annot_text and bari_annot==0:
        d['intersections'].text(bari_annot-0.25,y,f"{y:.1f}%",
                                ha='left',va='bottom',color="#f55f5f",zorder=10)
    
#     if intersections_ylabel    
#     if not post_fun is None: post_fun(ax['intersections'])
    return d


def plot_enrichment(
    data: pd.DataFrame,
    x: str,
    y: str,
    s: str,
    hue='Q',
    xlabel=None,
    ylabel='significance\n(-log10(Q))',
    size: int=None,
    color: str=None,
    annots_side: int=5,
    annots_side_labels=None,
    coff_fdr: float=None,
    xlim: tuple=None,
    xlim_off: float=0.2,
    ylim: tuple=None,
    ax: plt.Axes=None,
    break_pt: int=25,
    annot_coff_fdr: bool=False,
    kws_annot: dict=dict(
                    loc='right',
                    # annot_count_max=5,
                    offx3=0.15,
                    ),
    returns='ax',
    **kwargs
    ) -> plt.Axes:
    """Plot enrichment stats.

    Args:
        data (pd.DataFrame): input data.
        x (str): x column.
        y (str): y column.
        s (str): size column.
        size (int, optional): size of the points. Defaults to None.
        color (str, optional): color of the points. Defaults to None.
        annots_side (int, optional): how many labels to show on side. Defaults to 5.
        coff_fdr (float, optional): FDR cutoff. Defaults to None.
        xlim (tuple, optional): x-axis limits. Defaults to None.
        xlim_off (float, optional): x-offset on limits. Defaults to 0.2.
        ylim (tuple, optional): y-axis limits. Defaults to None.
        ax (plt.Axes, optional): `plt.Axes` object. Defaults to None.
        break_pt (int, optional): break point ('\n') for the labels. Defaults to 25.
        annot_coff_fdr (bool, optional): show FDR cutoff. Defaults to False.
        kws_annot (dict, optional): parameters provided to the `annot_side` function. Defaults to dict( loc='right', annot_count_max=5, offx3=0.15, ).

    Keyword Args:
        kwargs: parameters provided to the `sns.scatterplot` function. 

    Returns:
        plt.Axes: `plt.Axes` object.
    """
    if coff_fdr is None: 
        coff_fdr=1
    if any(data[y]==0):
        logging.warning(f"found {sum(data[y]==0)} zeros in among y values; replaced them with the next minimum which is {data[y].replace(0,np.nan).min()}.")
        data[y]=data[y].replace(0,data[y].replace(0,np.nan).min())
    from roux.stat.transform import log_pval
    # if y.startswith('P '):
    data[ylabel]=data[y].apply(log_pval)
    data[hue]=pd.cut(x=data[y],
                        bins=[
                            # data['P (FE test, FDR corrected)'].min(),
                            0,0.01,0.05,coff_fdr],
                        right=False,
                      )
    y=ylabel
    if not size is None:
        if not data[size].dtype == 'category':
            data[size]=pd.qcut(data[size],
                            q=3,
                            duplicates='drop')
        data=data.sort_values(size,ascending=False)
        data[size]=data[size].apply(lambda x: f"({x.left:.0f}, {x.right:.0f}]")
    if ax is None:
        fig,ax=plt.subplots()#(figsize=[1.5,4])
    sns.scatterplot(
                    data=data,
                    x=x,y=y,
                    size=size if not size is None else None,
                    size_order=data[size].unique() if not size is None else None,
                    hue=hue,
                    # color=color,
                    zorder=2,
                    ax=ax,
                    **kwargs,
    )
    if not size is None:
        ax.legend(loc='upper left',
                  bbox_to_anchor=(1.1, 0.1),
                 # title=size,
                  frameon=True,
        #          nrow=3,
                  ncol=2,)
    if xlim is None:
        ax=set_axlims(ax,off=xlim_off,axes=['x'])
    else:
        ax.set(xlim=xlim)
    # if ylim is None:
    #     ax.set(ylim=(log_pval(coff_fdr),ax.get_ylim()[1]),
    # #               xlim=(data[x].min(),data[x].max()),
    #           )
    # else:
    #     ax.set(ylim=ylim)        
    if annot_coff_fdr:
        ax.annotate(f"Q={coff_fdr}",
            xy=(ax.get_xlim()[0],log_pval(coff_fdr)), xycoords='data',
#             xytext=(-10,log_pval(coff_fdr)), textcoords='data',
            xytext=(0.01,0.1), textcoords='figure fraction',
            ha='left',va='top',
            arrowprops=dict(arrowstyle="->", color="0.5",
#                             shrinkA=5, shrinkB=5,
#                             patchA=None, patchB=None,
                            connectionstyle="arc3,rad=-0.3",
                            ),
            )
    from roux.viz.annot import annot_side
    ax=annot_side(
        ax=ax,
        df1=data.sort_values(y,ascending=False).head(annots_side) if annots_side_labels is None else \
            data.loc[data[s].isin(annots_side_labels),:],
        colx=x,
        coly=y,
        cols=s,
        break_pt=break_pt,
        # offymin=kws_annot['offymin'] if 'offymin' in kws_annot else if not size is None else 0,
        zorder=3,
        **kws_annot,
        )
    return locals()[returns]




from roux.global_imports import * 

## set
def set_(
    ax: plt.Axes,
    test: bool=False,
    **kws
    ) -> plt.Axes:
    """Ser many axis parameters.

    Args:
        ax (plt.Axes): `plt.Axes` object.
        test (bool, optional): test mode. Defaults to False.

    Keyword Args:
        kws: parameters provided to the `ax.set` function. 

    Returns:
        plt.Axes: `plt.Axes` object.
    """
    kws1={k:v for k,v in kws.items() if not isinstance(v,dict)}
    kws2={k:v for k,v in kws.items() if isinstance(v,dict)}    
    ax.set(**kws1)
    for k,v in kws2.items():
        getattr(ax,f"set_{k}")(**v)
    return ax

def set_ylabel(
    ax: plt.Axes,
    s: str=None,
    x: float=-0.1,
    y: float=1.02,
    xoff: float=0,
    yoff: float=0
    ) -> plt.Axes:
    """Set ylabel horizontal.

    Args:
        ax (plt.Axes): `plt.Axes` object.
        s (str, optional): ylabel. Defaults to None.
        x (float, optional): x position. Defaults to -0.1.
        y (float, optional): y position. Defaults to 1.02.
        xoff (float, optional): x offset. Defaults to 0.
        yoff (float, optional): y offset. Defaults to 0.

    Returns:
        plt.Axes: `plt.Axes` object.
    """
    if (not s is None):# and (ax.get_ylabel()=='' or ax.get_ylabel() is None):
        ax.set_ylabel(s)
    ax.set_ylabel(ax.get_ylabel(),rotation=0,ha='right',va='center')
    ax.yaxis.set_label_coords(x+xoff,y+yoff) 
    return ax
#     return set_label(x=x,y=y,s=ax.get_ylabel() if s is None else s,
#                                                        ha='right',va='bottom',ax=ax)

def rename_labels(ax,d1):
    ax.set_xlabel(replace_many(ax.get_xlabel(),d1,ignore=True))
    ax.set_ylabel(replace_many(ax.get_ylabel(),d1,ignore=True))
    ax.set_title(replace_many(ax.get_title(),d1,ignore=True))
    return ax

## ticklabels
def rename_ticklabels(
    ax: plt.Axes,
    axis: str,
    rename: dict=None,
    replace: dict=None,
    ignore: bool=False
    ) -> plt.Axes:
    """Rename the ticklabels.

    Args:
        ax (plt.Axes, optional): `plt.Axes` object. Defaults to None.
        axis (str): axis (x|y).
        rename (dict, optional): replace strings. Defaults to None.
        replace (dict, optional): replace sub-strings. Defaults to None.
        ignore (bool, optional): ignore warnings. Defaults to False.

    Raises:
        ValueError: either `rename` or `replace` should be provided.

    Returns:
        plt.Axes: `plt.Axes` object.
    """
    k=f"{axis}ticklabels"
    if not rename is None:
        _=getattr(ax,f"set_{k}")([rename[t.get_text()] for t in getattr(ax,f"get_{k}")()])
    elif not replace is None:
        _=getattr(ax,f"set_{k}")([replace_many(t.get_text(),replace,ignore=ignore) for t in getattr(ax,f"get_{k}")()])
    else:
        raise ValueError("either `rename` or `replace` should be provided.")
    return ax

def get_ticklabel2position(
    ax: plt.Axes,
    axis: str,
    ) -> plt.Axes:
    """Get positions of the ticklabels.

    Args:
        ax (plt.Axes): `plt.Axes` object.
        axis (str): axis (x|y).

    Returns:
        plt.Axes: `plt.Axes` object.
    """
    return dict(zip([t.get_text() for t in getattr(ax,f'get_{axis}ticklabels')()],
                  getattr(ax,f"{axis}axis").get_ticklocs()))

def set_ticklabels_color(
    ax: plt.Axes,
    ticklabel2color: dict,
    axis: str='y'
    ) -> plt.Axes:
    """Set colors to ticklabels.

    Args:
        ax (plt.Axes): `plt.Axes` object.
        ticklabel2color (dict): colors of the ticklabels.
        axis (str): axis (x|y).

    Returns:
        plt.Axes: `plt.Axes` object.
    """
    for tick in getattr(ax,f'get_{axis}ticklabels')():
        if tick.get_text() in ticklabel2color.keys():
            tick.set_color(ticklabel2color[tick.get_text()])
    return ax 
color_ticklabels=set_ticklabels_color

def format_ticklabels(
    ax: plt.Axes,
    axes: tuple=['x','y'],
    n: int=None,
    fmt: str=None,
    font: str=None,#'DejaVu Sans Mono',#"Monospace"
    ) -> plt.Axes:
    """format_ticklabels

    Args:
        ax (plt.Axes): `plt.Axes` object.
        axes (tuple, optional): axes. Defaults to ['x','y'].
        n (int, optional): number of ticks. Defaults to None.
        fmt (str, optional): format. Defaults to None.
        font (str, optional): font. Defaults to 'DejaVu Sans Mono'.

    Returns:
        plt.Axes: `plt.Axes` object.
        
    TODOs: 
        1. include color_ticklabels
    """
    if isinstance(n,int):
        n={'x':n,
           'y':n}
    if isinstance(fmt,str):
        fmt={'x':fmt,
           'y':fmt}
    for axis in axes:
        if not n is None:        
            getattr(ax,axis+'axis').set_major_locator(plt.MaxNLocator(n[axis]))
        if not fmt is None:
            getattr(ax,axis+'axis').set_major_formatter(plt.FormatStrFormatter(fmt[axis]))
        if not font is None:
            for tick in getattr(ax,f'get_{axis}ticklabels')():
                tick.set_fontname(font)
    return ax

## lims
def set_equallim(
    ax: plt.Axes,
    diagonal: bool=False,
    difference: float=None,
    format_ticks : bool=True,
    **kws_format_ticklabels,
    ) -> plt.Axes:
    """Set equal axis limits.

    Args:
        ax (plt.Axes): `plt.Axes` object.
        diagonal (bool, optional): show diagonal. Defaults to False.
        difference (float, optional): difference from . Defaults to None.

    Returns:
        plt.Axes: `plt.Axes` object.
    """
    min_,max_=np.min([ax.get_xlim()[0],ax.get_ylim()[0]]),np.max([ax.get_xlim()[1],ax.get_ylim()[1]])
    if diagonal:
        ax.plot([min_,max_],[min_,max_],':',color='gray',zorder=5)
    if not difference is None:
        off=np.sqrt(difference**2+difference**2)
        ax.plot([min_+off ,max_+off],[min_,max_],':',color='gray',zorder=5)        
        ax.plot([min_-off ,max_-off],[min_,max_],':',color='gray',zorder=5)      
    if format_ticks and len(ax.get_xticklabels())!=0:
        ax=format_ticklabels(ax,n=len(ax.get_xticklabels()),**kws_format_ticklabels)
        # logging.warning('format_ticklabels failed possibly because of shared axes (?).')
    ax.set_xlim(min_,max_)
    ax.set_ylim(min_,max_)
#     ax.set_xticks(ax.get_yticks())
    ax.set_aspect('equal', 'box')
    return ax

def get_axlims(ax: plt.Axes
    ) -> plt.Axes:
    """Get axis limits.

    Args:
        ax (plt.Axes): `plt.Axes` object.

    Returns:
        plt.Axes: `plt.Axes` object.
    """
    d1={}
    for axis in ['x','y']:
        d1[axis]={}
        d1[axis]['min'],d1[axis]['max']=getattr(ax,f'get_{axis}lim')()
        if d1[axis]['min'] > d1[axis]['max']:
            d1[axis]['min'],d1[axis]['max']=d1[axis]['max'],d1[axis]['min']
        d1[axis]['len']=abs(d1[axis]['min']-d1[axis]['max'])
    return d1


# axis limits    
def set_axlims(
    ax: plt.Axes,
    off: float,
    axes: list=['x','y']
    ) -> plt.Axes:
    """Set axis limits.

    Args:
        ax (plt.Axes): `plt.Axes` object.
        off (float): offset.
        axes (list, optional): axis name/s. Defaults to ['x','y'].

    Returns:
        plt.Axes: `plt.Axes` object.
    """
    logging.warning("prefer `ax.margins`")
    d1=get_axlims(ax)
    for k in axes:
        off_=(d1[k]['len'])*off
        if not getattr(ax,f"{k}axis").get_inverted():
            getattr(ax,f"set_{k}lim")(d1[k]['min']-off_,d1[k]['max']+off_)
        else:
            getattr(ax,f"set_{k}lim")(d1[k]['max']+off_,d1[k]['min']-off_)            
    return ax

def get_axlimsby_data(
    X: pd.Series,
    Y: pd.Series,
    off: float=0.2,
    equal: bool=False
    ) -> plt.Axes:
    """Infer axis limits from data.

    Args:
        X (pd.Series): x values.
        Y (pd.Series): y values.
        off (float, optional): offsets. Defaults to 0.2.
        equal (bool, optional): equal limits. Defaults to False.

    Returns:
        plt.Axes: `plt.Axes` object.
    """
    try:
        xmin=np.min(X)
        xmax=np.max(X)
    except:
        print(X)
    xlen=xmax-xmin
    ymin=np.min(Y)
    ymax=np.max(Y)
    ylen=ymax-ymin
    xlim=(xmin-off*xlen,xmax+off*xlen)
    ylim=(ymin-off*ylen,ymax+off*ylen)
    if not equal:
        return xlim,ylim
    else:
        lim=[np.min([xlim[0],ylim[0]]),np.max([xlim[1],ylim[1]])]
        return lim,lim

def split_ticklabels(
    ax: plt.Axes,
    axis='x',
    grouped=False,
    group_x=0.01,
    group_prefix=None,
    group_loc='left',
    # group_pad=0.02,
    group_colors=None,
    group_alpha=0.2,
    show_group_line=True,
    show_group_span=True,
    sep: str='-',
    pad_major=6,
    **kws,
    ) -> plt.Axes:
    """Split ticklabels into major and minor. Two minor ticks are created per major tick. 

    Args:
        ax (plt.Axes): `plt.Axes` object.
        sep (str, optional): separator within the tick labels. Defaults to ' '.

    Returns:
        plt.Axes: `plt.Axes` object.
    """
    ticklabels=getattr(ax,f'get_{axis}ticklabels')()
    import pandas as pd
    if not grouped:
        if axis=='y':logging.warning(f'axis={axis} is not tested.')        
        ticklabels_major=pd.unique(['\u2014\n'+s.get_text().split(sep)[0] for s in ticklabels])
        ticklabels_minor=[s.get_text().split(sep)[1] for s in ticklabels]

        ticks_minor=getattr(ax,f'get_{axis}ticks')()
        ticks_major=ticks_minor.reshape(int(len(ticks_minor)/2),2).mean(axis=1)
        _=getattr(ax,f'set_{axis}ticks')(ticks_major, minor=False)
        getattr(ax,f'set_{axis}ticklabels')(ticklabels_major,minor=False,**kws,)
        _=ax.set_xticks( ticks_minor, minor=True )
        getattr(ax,f'set_{axis}ticklabels')(ticklabels_minor,minor=True,**kws,)
        ax.tick_params(axis=axis, which='minor', bottom=True,pad=0)
        ax.tick_params(axis=axis, which='major', bottom=False,pad=pad_major)
    else:
        if axis=='x':logging.warning(f'axis={axis} is not tested.')
        from roux.lib.df import dict2df
        df0_=dict2df(get_ticklabel2position(ax=ax,axis=axis),
                   colkey=axis+'ticklabel',colvalue=axis)
        df0_[axis+'ticklabel major']=df0_[axis+'ticklabel'].str.split(sep,1,expand=True)[0]
        df0_[axis+'ticklabel minor']=df0_[axis+'ticklabel'].str.split(sep,1,expand=True)[1]
        df_=(df0_
        .groupby(axis+'ticklabel major')
        .agg({axis:[min,max,len],})
        .rd.flatten_columns()
        )
        axlims=get_axlims(ax)
        # if group_loc=='left':
        #     group_x=group_x
        #     group_x=axlims[axis]['min']-(axlims[axis]['len']*group_pad)
        #     group_xlabel=axlims[axis]['min']-(axlims[axis]['len']*group_pad-0.1)
        # elif group_loc=='right':
        #     group_x=axlims[axis]['max']+(axlims[axis]['len']*group_pad)
        #     group_xlabel=axlims[axis]['max']+(axlims[axis]['len']*group_pad+0.1)
        # print(axlims[axis]['min']-(group_pad*5.5))
        if show_group_span:
            # print(axlims[axis]['min']-(group_pad*5.5))
            # print(group_x)
            df_.apply(lambda x: ax.axhspan(
            # xmin=(group_x/axlims[axis]['len'])*0.5,
            # xmin=axlims[axis]['min']-(group_pad*5.5),
            xmin=group_x,
            xmax=axlims[axis]['min'], 
            ymin=x[axis+' min']-0.5,
            ymax=x[axis+' max']+0.5, 
            transform="axes",
            clip_on=False,
            zorder=0,
            color=None if group_colors is None else group_colors[x.name],
            edgecolor='none',
            alpha=group_alpha,
            ),
            axis=1)
        if show_group_line:
            df_.apply(lambda x: ax.plot(
                [group_x,group_x],
                [x[axis+' min']-0.2,x[axis+' max']+0.2],
                clip_on=False,
                lw=0.5,
                color='k',
                # transform="axes",
                ),
                axis=1)
        from roux.lib.set import get_alt
        import matplotlib.transforms as transforms
        df_.apply(lambda x: ax.text(
                                    x=group_x,#label,
                                    y=np.mean([x[axis+' min'],x[axis+' max']]),
                                    s=(group_prefix+'\n' if not group_prefix is None else '')+f"{x.name}".replace(' ','\n')+'\n'+f"(n={int(x[axis+' len'])})",
                                    color='k',
                                    # ha=get_alt(['left','right'],group_loc),
                                    ha=group_loc,
                                    va='center',
                                    transform=transforms.blended_transform_factory(ax.transAxes,ax.transData),
                                    # transform=ax.transAxes,
                                   ),axis=1)
        getattr(ax,f'set_{axis}ticklabels')([s.get_text().split(sep,1)[1] for s in ticklabels],
                                            **kws,)        
    return ax

def set_grids(
    ax: plt.Axes,
    axis: str=None
    ) -> plt.Axes:
    """Show grids.

    Args:
        ax (plt.Axes): `plt.Axes` object.
        axis (str, optional): axis name. Defaults to None.

    Returns:
        plt.Axes: `plt.Axes` object.
    """
    w,h=ax.figure.get_size_inches()
    if w/h>=1.1 or axis=='y' or axis=='both':
        ax.set_axisbelow(True)
        ax.yaxis.grid(color='gray', linestyle='dashed')
    if w/h<=0.9 or axis=='x' or axis=='both':
        ax.set_axisbelow(True)
        ax.xaxis.grid(color='gray', linestyle='dashed')
    return ax

## legends
def rename_legends(
    ax: plt.Axes,
    replaces: dict,
    **kws_legend
    ) -> plt.Axes:
    """Rename legends.

    Args:
        ax (plt.Axes): `plt.Axes` object.
        replaces (dict): _description_

    Returns:
        plt.Axes: `plt.Axes` object.
    """
    handles, labels = ax.get_legend_handles_labels()
    labels=[str(s) for s in labels]
    if len(set(labels) - set(replaces.keys()))==0:
        labels=[replaces[s] for s in labels]
    else:
        labels=[replacemany(s,replaces) for s in labels]
    return ax.legend(handles=handles,labels=labels,
                     title=ax.get_legend().get_title().get_text(),
                     **kws_legend)

def append_legends(
    ax: plt.Axes,
    labels: list,
    handles: list,
    **kws
    ) -> plt.Axes:
    """Append to legends.

    Args:
        ax (plt.Axes): `plt.Axes` object.
        labels (list): labels. 
        handles (list): handles.

    Returns:
        plt.Axes: `plt.Axes` object.
    """
    h1, l1 = ax.get_legend_handles_labels()
    print(l1)
    ax.legend(handles=h1+handles,
              labels=l1+labels,
              **kws)
    return ax

## legend related stuff: also includes colormaps

def sort_legends(
    ax: plt.Axes,
    sort_order: list=None,
    **kws
    ) -> plt.Axes:
    """Sort or filter legends.

    Args:
        ax (plt.Axes): `plt.Axes` object.
        sort_order (list, optional): order of legends. Defaults to None.

    Returns:
        plt.Axes: `plt.Axes` object.
        
    Notes:
        1. Filter the legends by providing the indices of the legends to keep.
    """
    handles, labels = ax.get_legend_handles_labels()
    # sort both labels and handles by labels
    if sort_order is None:
        handles,labels = zip(*sorted(zip(handles,labels), key=lambda t: t[1]))
    else:
        if all([isinstance(i,str) for i in sort_order]):
            sort_order=[labels.index(s) for s in sort_order]
        if not all([isinstance(i,int) for i in sort_order]):
            raise ValueError("sort_order should contain all integers")
        handles,labels =[handles[idx] for idx in sort_order],[labels[idx] for idx in sort_order]
        # print(handles,labels)
    return ax.legend(handles, labels,**kws)

def drop_duplicate_legend(ax,**kws): return sort_legends(ax=ax,sort_order=None,**kws)

def reset_legend_colors(ax):
    """Reset legend colors.

    Args:
        ax (plt.Axes): `plt.Axes` object.

    Returns:
        plt.Axes: `plt.Axes` object.
    """
    leg=plt.legend()
    for lh in leg.legendHandles: 
        lh.set_alpha(1)
#         lh._legmarker.set_alpha(1)
    return ax

def set_legends_merged(axs):
    """Reset legend colors.

    Args:
        axs (list): list of `plt.Axes` objects.

    Returns:
        plt.Axes: first `plt.Axes` object in the list.
    """
    df_=pd.concat([pd.DataFrame(ax.get_legend_handles_labels()[::-1]).T for ax in axs],
             axis=0)
    df_['fc']=df_[1].apply(lambda x: x.get_fc())
    df_=df_.log.drop_duplicates(subset=[0,'fc'])
    if df_[0].duplicated().any(): logging.error("duplicate legend labels")
    return axs[1].legend(handles=df_[1].tolist(), labels=df_[0].tolist(),
                       bbox_to_anchor=[-0.2,0],loc=2,frameon=True).get_frame().set_edgecolor((0.95,0.95,0.95))

def set_legend_custom(
    ax: plt.Axes,
    legend2param: dict,
    param: str='color',
    lw: float=1,
    marker: str='o',
    markerfacecolor: bool=True,
    size: float=10,
    color: str='k',
    linestyle: str='',
    title_ha: str='center',
    frameon: bool=True,
    **kws
    ) -> plt.Axes:
    """Set custom legends.

    Args:
        ax (plt.Axes): `plt.Axes` object.
        legend2param (dict): legend name to parameter to change e.g. name of the color. 
        param (str, optional): parameter to change. Defaults to 'color'.
        lw (float, optional): line width. Defaults to 1.
        marker (str, optional): marker type. Defaults to 'o'.
        markerfacecolor (bool, optional): marker face color. Defaults to True.
        size (float, optional): size of the markers. Defaults to 10.
        color (str, optional): color of the markers. Defaults to 'k'.
        linestyle (str, optional): line style. Defaults to ''.
        title_ha (str, optional): title horizontal alignment. Defaults to 'center'.
        frameon (bool, optional): show frame. Defaults to True.

    Returns:
        plt.Axes: `plt.Axes` object.
        
    TODOs:
        1. differnet number of points for eachh entry
        
            from matplotlib.legend_handler import HandlerTuple
            l1, = plt.plot(-1, -1, lw=0, marker="o",
                        markerfacecolor='k', markeredgecolor='k')
            l2, = plt.plot(-0.5, -1, lw=0, marker="o",
                        markerfacecolor="none", markeredgecolor='k')
            plt.legend([(l1,), (l1, l2)], ["test 1", "test 2"], 
                    handler_map={tuple: HandlerTuple(2)}
                    )
    
    References: 
        https://matplotlib.org/stable/api/markers_api.html
        http://www.cis.jhu.edu/~shanest/mpt/js/mathjax/mathjax-dev/fonts/Tables/STIX/STIX/All/All.html
    """
    from matplotlib.lines import Line2D
    legend_elements=[Line2D([0], [0],
                       marker=marker,
                       color=color if param!='color' else legend2param[k],
                       markeredgecolor=(color if param!='color' else legend2param[k]), 
                       markerfacecolor=(color if param!='color' else legend2param[k]) if not markerfacecolor is None else 'none',
                       markersize=(size if param!='size' else legend2param[k]),
                       label=k,
                       lw=(lw if param!='lw' else legend2param[k]),
                       linestyle=linestyle if param!='lw' else '-',
                      ) for k in legend2param]   
    o1=ax.legend(handles=legend_elements,frameon=frameon,
              **kws)
    o1._legend_box.align=title_ha
    o1.get_frame().set_edgecolor((0.95,0.95,0.95))
    return ax

## line round
def get_line_cap_length(
    ax: plt.Axes,
    linewidth: float
    ) -> plt.Axes:
    """Get the line cap length.

    Args:
        ax (plt.Axes): `plt.Axes` object
        linewidth (float): width of the line.

    Returns:
        plt.Axes: `plt.Axes` object
    """
    radius = linewidth / 2
    ppd = 72. / ax.figure.dpi  # points per dot
    trans = ax.transData.inverted().transform
    x_radius = ((trans((radius / ppd, 0)) - trans((0, 0))))[0]
    y_radius = ((trans((0, radius / ppd)) - trans((0, 0))))[1]
    return x_radius,y_radius


# shape
def get_subplot_dimentions(ax=None):
    """Calculate the aspect ratio of `plt.Axes`.

    Args:
        ax (plt.Axes): `plt.Axes` object

    Returns:
        plt.Axes: `plt.Axes` object

    References: 
        https://github.com/matplotlib/matplotlib/issues/8013#issuecomment-285472404    
    """
    if ax is None:
        ax = plt.gca()
    fig = ax.figure

    ll, ur = ax.get_position() * fig.get_size_inches()
    width, height = ur - ll
    return width, height,height / width

# colorbar
def set_colorbar(
    fig: object,
    ax: plt.Axes,
    ax_pc: plt.Axes,
    label: str,
    bbox_to_anchor: tuple=(0.05, 0.5, 1, 0.45),
    orientation: str="vertical",
    ):
    """Set colorbar.

    Args:
        fig (object): figure object.
        ax (plt.Axes): `plt.Axes` object.
        ax_pc (plt.Axes): `plt.Axes` object for the colorbar.
        label (str): label
        bbox_to_anchor (tuple, optional): location. Defaults to (0.05, 0.5, 1, 0.45).
        orientation (str, optional): orientation. Defaults to "vertical".

    Returns:
        figure object.
    """
    from mpl_toolkits.axes_grid1.inset_locator import inset_axes
    if orientation=="vertical":
        width,height="5%","50%"
    else:
        width,height="50%","5%"        
    axins = inset_axes(ax,
                       width=width,  # width = 5% of parent_bbox width
                       height=height,  # height : 50%
                       loc=2,
                       bbox_to_anchor=bbox_to_anchor,
                       bbox_transform=ax.transAxes,
                       borderpad=0,
                       )
    fig.colorbar(ax_pc, cax=axins,
                 label=label,orientation=orientation,)
    return fig

def set_colorbar_label(
    ax: plt.Axes,
    label: str
    ) -> plt.Axes:
    """Find colorbar and set label for it.

    Args:
        ax (plt.Axes): `plt.Axes` object.
        label (str): label.

    Returns:
        plt.Axes: `plt.Axes` object.
    """
    for a in ax.figure.get_axes()[::-1]:
        if a.properties()['label']=='<colorbar>':
            if hasattr(a,'set_ylabel'):
                a.set_ylabel(label)
                break
    return ax
    


from roux.lib.df import *

def log_likelihood(y_true: list, y_pred: list) -> float:
    """Log likelihood.

    Args:
        y_true (list): True
        y_pred (list): Predicted.

    Returns:
        float: log likelihood

    Reference: 
        1. https://github.com/saezlab/protein_attenuation/blob/6c1e81af37d72ef09835ee287f63b000c7c6663c/src/protein_attenuation/utils.py
    """
    n = len(y_true)
    ssr = np.power(y_true - y_pred, 2).sum()
    var = ssr / n

    l = np.longfloat(1 / (np.sqrt(2 * np.pi * var))) ** n * np.exp(-(np.power(y_true - y_pred, 2) / (2 * var)).sum())
    ln_l = np.log(l)

    return ln_l

def f_statistic(y_true, y_pred, n, p):
    """F-statistic.

    Args:
        y_true (list): True
        y_pred (list): Predicted.

    Returns:
        float: F-statistic

    Reference: 
        1. https://github.com/saezlab/protein_attenuation/blob/6c1e81af37d72ef09835ee287f63b000c7c6663c/src/protein_attenuation/utils.py
    """
    msm = np.power(y_pred - y_true.mean(), 2).sum() / p
    mse = np.power(y_true - y_pred, 2).sum() / (n - p - 1)

    f = msm / mse

    f_pval = stats.f.sf(f, p, n - p - 1)

    return f, f_pval
    
def compare_bools_jaccard(x,y):
    """Compare bools in terms of the jaccard index.

    Args:
        x (list): list of bools.
        y (list): list of bools.

    Returns:
        float: jaccard index.
    """
    x = np.asarray(x, np.bool)
    y = np.asarray(y, np.bool)
    return np.double(np.bitwise_and(x, y).sum()) / np.double(np.bitwise_or(x, y).sum())

def compare_bools_jaccard_df(df: pd.DataFrame) -> pd.DataFrame:
    """Pairwise compare bools in terms of the jaccard index.

    Args:
        df (DataFrame): dataframe with boolean columns.

    Returns:
        DataFrame: matrix with comparisons between the columns.
    """
    from roux.stat.binary import compare_bools_jaccard
    dmetrics=pd.DataFrame(index=df.columns.tolist(),columns=df.columns.tolist())
    for c1i,c1 in enumerate(df.columns):
        for c2i,c2 in enumerate(df.columns):
            if c1i>c2i:
                dmetrics.loc[c1,c2]=compare_bools_jaccard(df.dropna(subset=[c1,c2])[c1],df.dropna(subset=[c1,c2])[c2])
            elif c1i==c2i:
                dmetrics.loc[c1,c2]=1
    for c1i,c1 in enumerate(df.columns):
        for c2i,c2 in enumerate(df.columns):
            if c1i<c2i:
                dmetrics.loc[c1,c2]=dmetrics.loc[c2,c1]
    return dmetrics

def classify_bools(l: list) -> str:
    """Classify bools.

    Args:
        l (list): list of bools

    Returns:
        str: classification.
    """
    return 'both' if all(l) else 'either' if any(l) else 'neither'

## agg
def frac(x: list) -> float:
    """Fraction.

    Args:
        x (list): list of bools.

    Returns:
        float: fraction of True values.
    """
    return (sum(x)/len(x))
def perc(x: list) -> float:
    """Percentage.

    Args:
        x (list): list of bools.

    Returns:
        float: Percentage of the True values
    """
    return frac(x)*100

## confusion_matrix
def get_stats_confusion_matrix(df_: pd.DataFrame) -> pd.DataFrame:
    """Get stats confusion matrix.

    Args:
        df_ (DataFrame): Confusion matrix.

    Returns:
        DataFrame: stats.
    """
    d0={}
    d0['TP']=df_.loc[True,True]
    d0['TN']=df_.loc[False,False]
    d0['FP']=df_.loc[False,True]
    d0['FN']=df_.loc[True,False]
    # Sensitivity, hit rate, recall, or true positive rate
    d0['TPR'] = d0['TP']/(d0['TP']+d0['FN'])
    # Specificity or true negative rate
    d0['TNR'] = d0['TN']/(d0['TN']+d0['FP']) 
    # Precision or positive predictive value
    d0['PPV'] = d0['TP']/(d0['TP']+d0['FP'])
    # Negative predictive value
    d0['NPV'] = d0['TN']/(d0['TN']+d0['FN'])
    # Fall out or false positive rate
    d0['FPR'] = d0['FP']/(d0['FP']+d0['TN'])
    # False negative rate
    d0['FNR'] = d0['FN']/(d0['TP']+d0['FN'])
    # False discovery rate
    d0['FDR'] = d0['FP']/(d0['TP']+d0['FP'])
    # Overall accuracy
    d0['ACC'] = (d0['TP']+d0['TN'])/(d0['TP']+d0['FP']+d0['FN']+d0['TN'])
    df1=pd.Series(d0).to_frame('value')
    df1.index.name='variable'
    df1=df1.reset_index()
    return df1


import logging
import pandas as pd
from icecream import ic as info

## metadata to process the data
def exclude_items(
    df1 : pd.DataFrame,
    metadata: dict,
    ) -> pd.DataFrame:
    """Exclude items from the table with the workflow info.

    Args:
        df1 (pd.DataFrame): input table.
        metadata (dict): metadata of the repository.

    Returns:
        pd.DataFrame: output.
    """
    for c in df1:
        if c in metadata['exclude']:
            # info(c)
            shape0=df1.shape
            df1=df1.loc[~(df1[c].isin(metadata['exclude'][c])),:]
            if shape0!=df1.shape:
                logging.info(f"{c}:{df1[c].nunique()}")
    return df1



"""
End-point.
-> global_imports
"""
import logging
logger = logging.getLogger()
logger.setLevel(logging.INFO)
# temporary change for a context
# ref: https://docs.python.org/3/howto/logging-cookbook.html#using-a-context-manager-for-selective-logging
# class LoggingContext:
#     def __init__(self, logger, level=None, handler=None, close=True):
#         self.logger = logger
#         self.level = level
#         self.handler = handler
#         self.close = close

#     def __enter__(self):
#         if self.level is not None:
#             self.old_level = self.logger.level
#             self.logger.setLevel(self.level)
#         if self.handler:
#             self.logger.addHandler(self.handler)

#     def __exit__(self, et, ev, tb):
#         if self.level is not None:
#             self.logger.setLevel(self.old_level)
#         if self.handler:
#             self.logger.removeHandler(self.handler)
#         if self.handler and self.close:
#             self.handler.close()
# with LoggingContext(logger, level=logging.ERROR):
#     logger.debug('3. This should appear once on stderr.')

# recepies
from pathlib import Path
from roux.lib.str import get_bracket, replace_many,get_suffix,get_prefix
from roux.lib.dict import *
from roux.lib.set import *
from roux.lib.text import read,read_lines
from roux.lib.io import * #io_df -> io_dfs -> io_files
from roux.lib.dict import to_dict # to replace io_df to_dict
from roux.workflow.io import read_metadata, to_diff_notebooks, clear_variables, clear_dataframes
from roux.workflow.df import *

# defaults
# from functools import partialmethod
# pd.DataFrame.head = partialmethod(pd.DataFrame.head, n=1)
# pd.set_option('display.max_rows', 2)

# stats    
import scipy as sc
from roux.stat.binary import perc

# paths
pwd=abspath('.')

# plots
import matplotlib.pyplot as plt
import seaborn as sns
plt.style.use('ggplot')
plt.rcParams['pdf.fonttype'] = 42
plt.rcParams['ps.fonttype'] = 42
plt.rcParams['font.family'] = 'Myriad Pro'
plt.rcParams['figure.figsize'] = (3, 3)
plt.rcParams['axes.facecolor']='none'
plt.rcParams['axes.edgecolor']='k'
# plt.rcParams['axes.formatter.use_mathtext'] = True
plt.rcParams['axes.formatter.limits'] = -3, 3
plt.rcParams['axes.formatter.min_exponent'] = 3
plt.rcParams['legend.frameon']=True
from cycler import cycler
plt.rcParams['axes.prop_cycle']= cycler('color',[
    '#50AADC',#blue
    "#D3DDDC",#gray
    '#F1D929',#yellow
    "#f55f5f",#red
    "#046C9A",#blue
    "#00A08A", "#F2AD00", "#F98400", "#5BBCD6", "#ECCBAE", "#D69C4E", "#ABDDDE", "#000000"])
from roux.viz.colors import get_colors_default
# plt.rcParams['xtick.color']=[0.95,0.95,0.95]
plt.rc('grid', lw=0.2,linestyle="-", color=[0.98,0.98,0.98])
plt.rc('axes', axisbelow=True)
plt.rc('axes', unicode_minus=False)
plt.rcParams['axes.labelcolor'] = 'k'
sns.set_context('notebook') # paper < notebook < talk < poster
from roux.viz.figure import *
from roux.viz.io import log_code,get_plot_inputs,to_plot,read_plot#*
from roux.viz.ax_ import *
from roux.viz.annot import *

from tqdm import tqdm#,notebook
# from roux.lib.sys import is_interactive_notebook
if not is_interactive_notebook:
    from IPython import get_ipython
    logging.info("log_notebookp=f'log_notebook.log';open(log_notebookp, 'w').close();get_ipython().run_line_magic('logstart','{log_notebookp} over')")
    tqdm.pandas()
else:
    from tqdm import notebook
    notebook.tqdm().pandas()
    ## display vector graphics in jupyter
    # if not get_ipython() is None:
    #     get_ipython().run_line_magic('config', "InlineBackend.figure_formats = ['svg']")

## session info
import watermark.watermark as watermark
logging.info(watermark(python=True)+watermark(iversions=True,globals_=globals()))

## parallel processing
from pandarallel import pandarallel
pandarallel.initialize(nb_workers=6,progress_bar=True)
# logging.info("pandarallel.initialize(nb_workers=4,progress_bar=True)")

# metadata
# if not 'metadata' in globals():
#     metadata=read_metadata()


import pandas as pd
# stats 
def perc_label(a,b=None,bracket=True): 
    from roux.lib.str import num2str
    if b is None:
        b=len(a)
        a=sum(a)
    return f"{(a/b)*100:.0f}%"+(f" ({num2str(a)}/{num2str(b)})" if bracket else "")

def pval2annot(pval,
               alternative=None,
               alpha=None,
               fmt='*',
               power=True,
               linebreak=False,
               prefix='P',
               q=False,
              ):
    """
    fmt: *|<|'num'    
    """
    if alternative is None and alpha is None:
        raise ValueError('both alternative and alpha are None')
    if alpha is None:
        alpha=0.025 if alternative=='two-sided' else 0.05
    if pd.isnull(pval):
        annot= ''
    elif pval < 0.0001:
        annot= "****" if fmt=='*' else \
        f"P<\n{0.0001:.0e}" if fmt=='<' else \
        f"P={pval:.1g}" if len(f"P={pval:.1g}")<6 else \
        f"P=\n{pval:.1g}"  if not linebreak else \
        f"P={pval:.1g}"
    elif (pval < 0.001):
        annot= "***"  if fmt=='*' else \
        f"P<\n{0.001:.0e}" if fmt=='<' else \
        f"P={pval:.1g}" if len(f"P={pval:.1g}")<6 else \
        f"P=\n{pval:.1g}" if not linebreak else \
        f"P={pval:.1g}"
    elif (pval < 0.01):
        annot= "**" if fmt=='*' else \
        f"P<\n{0.01:.0e}" if fmt=='<' else \
        f"P={pval:.1g}" if len(f"P={pval:.1g}")<6 else \
        f"P=\n{pval:.1g}" if not linebreak else \
        f"P={pval:.1g}"
    elif (pval < alpha):
        annot= "*" if fmt=='*' else \
        f"P<\n{alpha}" if fmt=='<' else \
        f"P={pval:.1g}" if len(f"P={pval:.1g}")<6 else \
        f"P=\n{pval:.1g}" if not linebreak else \
        f"P={pval:.1g}"
    else:
        annot= "ns" if fmt=='*' else \
        f"P={pval:.1g}" if len(f"P={pval:.1g}")<6 else \
        f"P=\n{pval:.1g}" if not linebreak else \
        f"P={pval:.1g}"
    annot=annot if linebreak else annot.replace('\n','')
    if prefix!='P':
        annot=annot.replace('P',prefix if not q else 'q')
    if power and 'e' in annot:
        annot=annot.replace('e-0','e-').replace('e','x$10^{')+'}$'
    return annot 


import matplotlib.pyplot as plt

## set subplots
def get_subplots(
    nrows: int,
    ncols: int,
    total: int=None
    ) -> list:
    """Get subplots.

    Args:
        nrows (int): number of rows.
        ncols (int): number of columns.
        total (int, optional): total subplots. Defaults to None.

    Returns:
        list: list of `plt.Axes` objects.
    """
    idxs=list(itertools.product(range(nrows),range(ncols)))
    if not total is None:
        idxs=idxs[:total]
    print(idxs)
    return [plt.subplot2grid([nrows,ncols],idx,1,1) for idx in idxs]

def labelplots(
    fig,
    axes: list=None,
    labels: list=None,
    xoff: float=0,
    yoff: float=0,
    params_alignment: dict={},
    size:float = 20,
    va:str = 'bottom',
    ha:str = 'right',
    test: bool=False,
    **kws,
    ):
    """Label (sub)plots.

    Args:
        fig : `plt.figure` object. 
        axes (_type_): list of `plt.Axes` objects.
        xoff (int, optional): x offset. Defaults to 0.
        yoff (int, optional): y offset. Defaults to 0.
        params_alignment (dict, optional): alignment parameters. Defaults to {}.
        params_text (dict, optional): parameters provided to `plt.text`. Defaults to {'size':20,'va':'bottom', 'ha':'right' }.
        test (bool, optional): test mode. Defaults to False.
    """
    if axes is None:
        axes=fig.axes
    if labels is None:
        import string
        labels=string.ascii_uppercase[:len(axes)]
    else:
        assert len(axes)==len(labels) 
    label2ax=dict(zip(labels,axes))
    axi2xy={}
    for axi,label in enumerate(label2ax.keys()):
        ax=label2ax[label]
        axi2xy[axi]=ax.get_position(original=True).xmin+xoff,ax.get_position(original=False).ymax+yoff
    for pair in params_alignment:
        axi2xy[pair[1]]=[axi2xy[pair[0 if 'x' in params_alignment[pair] else 1]][0],
                         axi2xy[pair[0 if 'y' in params_alignment[pair] else 1]][1]]
    for axi,label in enumerate(label2ax.keys()):
        label2ax[label].text(*axi2xy[axi],f"{label}",
                             size=size,
                             ha=ha,
                             va=va,
                             transform=fig.transFigure,
                             **kws)    




import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import logging
from roux.lib.str import *

# redirects
from roux.stat.io import perc_label,pval2annot

# labels 
def set_label(
    s: str,
    ax: plt.Axes,
    x: float= 0,
    y: float= 0,
    ha: str='left',
    va: str='top',
    loc=None,
    off_loc=0.01,
    title: bool=False,
    **kws,
    ) -> plt.Axes:
    """Set label on a plot.

    Args:
        x (float): x position.
        y (float): y position.
        s (str): label.
        ax (plt.Axes): `plt.Axes` object.
        ha (str, optional): horizontal alignment. Defaults to 'left'.
        va (str, optional): vertical alignment. Defaults to 'top'.
        loc (int, optional): location of the label. 1:'upper right', 2:'upper left', 3:'lower left':3, 4:'lower right'
        offs_loc (tuple,optional): x and y location offsets.
        title (bool, optional): set as title. Defaults to False.
        
    Returns:
        plt.Axes: `plt.Axes` object.
    """
    if title:
        ax.set_title(s,**kws)
    elif not loc is None:
        if loc==1 or loc=='upper right':
            x=1-off_loc
            y=1-off_loc
            ha='right'
            va='top'
        elif loc==2 or loc=='upper left':
            x=0+off_loc
            y=1-off_loc
            ha='left'
            va='top'
        elif loc==3 or loc=='lower left':
            x=0+off_loc
            y=0+off_loc
            ha='left'
            va='bottom'            
        elif loc==4 or loc=='lower right':
            x=1-off_loc
            y=0+off_loc
            ha='right'
            va='bottom'            
        else:
            raise ValueError(loc)
    ax.text(s=s,transform=ax.transAxes,
            x=x,y=y,ha=ha,va=va,
            **kws)
    return ax

def annot_side(
    ax: plt.Axes,
    df1: pd.DataFrame,
    colx: str,
    coly: str,
    cols: str=None,
    hue: str=None,
    loc: str='right',
    scatter=False,
    lines=True,
    text=True,
    invert_xaxis: bool=False, 
    offx3: float=0.15,
    offymin: float=0.1,
    offymax: float=0.9,
    break_pt: int=25,
    length_axhline: float=3,
    va: str='bottom',
    zorder: int=1,
    color: str='gray',
    kws_line: dict={},
    kws_scatter: dict={'zorder':2,'alpha':0.75,'marker':'|','s':100},
    **kws_text,
    ) -> plt.Axes:
    """Annot elements of the plots on the of the side plot.

    Args:
        df1 (pd.DataFrame): input data
        colx (str): column with x values.
        coly (str): column with y values.
        cols (str): column with labels.
        hue (str): column with colors of the labels.
        ax (plt.Axes, optional): `plt.Axes` object. Defaults to None.
        loc (str, optional): location. Defaults to 'right'.
        invert_xaxis (bool, optional): invert xaxis. Defaults to False.
        offx3 (float, optional): x-offset for bend position of the arrow. Defaults to 0.15.
        offymin (float, optional): x-offset minimum. Defaults to 0.1.
        offymax (float, optional): x-offset maximum. Defaults to 0.9.
        break_pt (int, optional): break point of the labels. Defaults to 25.
        length_axhline (float, optional): length of the horizontal line i.e. the "underline". Defaults to 3.
        zorder (int, optional): z-order. Defaults to 1.
        color (str, optional): color of the line. Defaults to 'gray'.
        kws_line (dict, optional): parameters for formatting the line. Defaults to {}.

    Keyword Args:
        kws: parameters provided to the `ax.text` function. 

    Returns:
        plt.Axes: `plt.Axes` object.
    """
    if len(df1)==0: 
        logging.warning("annot_side: no data found")
        return
    if isinstance(colx,float):
        df1['colx']=colx
        colx='colx'
    if isinstance(coly,float):
        df1['coly']=coly
        coly='coly'
    # assert not 'y' in df1, 'table should not contain a column named `y`'
    df1=df1.sort_values(coly if loc!='top' else colx,ascending=True)
    from roux.viz.ax_ import get_axlims
    d1=get_axlims(ax)
    # if loc=='top', annotations x is y and y is x
    df1['y']=np.linspace(d1['y' if loc!='top' else 'x']['min']+((d1['y' if loc!='top' else 'x']['len'])*offymin),
                        d1['y' if loc!='top' else 'x']['max']*offymax,
                        len(df1))
    x2=d1['x'if loc!='top' else 'y']['min' if not invert_xaxis else 'max'] if loc=='left' else d1['x'if loc!='top' else 'y']['max' if not invert_xaxis else 'min']
    x3=d1['x']['min']-(d1['x']['len']*offx3) if (loc=='left' and not invert_xaxis) else \
        d1['y']['max']+(d1['y']['len']*offx3) if loc=='top' else \
        d1['x']['max']+(d1['x']['len']*offx3)
    # print(x2,x3)
    # line#1
    # print(df1.loc[:,[colx,coly,'y']].iloc[0,:])
    if lines:
        df1.apply(lambda x: ax.plot([x[colx],x2] if loc!='top' else [x[colx], x['y']],
                                   [x[coly],x['y']] if loc!='top' else [x[coly], x2],
                                   color=color,lw=1,**kws_line,
                                   zorder=zorder,
                                   ),axis=1)
    if scatter:
        df1.plot.scatter(x=colx,y=coly,ax=ax,
                        **kws_scatter)
        
    ## text
    if text:
        df1.apply(lambda x: ax.text(x3 if loc!='top' else x['y'],
                                    x['y'] if loc!='top' else x3,
                                    linebreaker(x[cols],break_pt=break_pt,),
                                    ha='right' if loc=='left' else 'center' if loc=='top' else 'left',
                                    va=va,
                                    color=x[hue] if not hue is None else 'k',
                                    # **{k:v for k,v in kws_text.items() if not (k==color and not hue is None)},
                                    rotation=0  if loc!='top' else 90,
                                    **kws_text,
                                  zorder=2),axis=1)
    # line #2
    if lines:
        if loc!='top':
            df1.apply(lambda x:ax.axhline(y = x['y'], 
                                         xmin=0 if loc=='left' else 1,
                                         xmax=0-(length_axhline-1)-offx3 if loc=='left' else length_axhline+offx3,
                                                 clip_on = False,color=color,lw=1,
                                        ),axis=1)
        else:
            df1.apply(lambda x:ax.axvline(x = x['y'], 
                                         ymin=0 if loc=='left' else 1,
                                         ymax=0-(length_axhline-1)-offx3 if loc=='left' else length_axhline+offx3,
                                                 clip_on = False,color=color,lw=1,
                                        ),axis=1)
    if loc=='left':
        ax.yaxis.tick_right()
        ax.yaxis.set_label_position("right")
    ax.set(xlim=[d1['x']['min'],d1['x']['max']],
           ylim=[d1['y']['min'],d1['y']['max']],
            )
    return ax

def annot_corners(
    ax : plt.Axes,
    df1 : pd.DataFrame,
    colx : str,
    coly : str,
    coltext : str,
    off : float=0.1,
    **kws,
    ) -> plt.Axes:
    """
    Annotate points above and below the diagonal.
    """
    df1['diff']=df1[coly]-df1[colx]
    # above diagonal
    df1['loc']=df1['diff'].apply(lambda x: 'above' if x>0 else 'below')
    from roux.viz.ax_ import get_axlims
    axlims=get_axlims(ax=ax)
    for loc,df1_ in df1.groupby('loc'):
        df1_=df1_.sort_values([colx,coly])        
        offx=axlims['x']['len']*off
        offy=axlims['y']['len']*off
        # upper
        df1_['x text' if loc=='above' else 'y text']=np.linspace(axlims['x' if loc=='above' else 'y']['min']+offx,
                                (axlims['x' if loc=='above' else 'y']['min']+axlims['x' if loc=='above' else 'y']['len']*0.5)-offx,
                                len(df1_))
        df1_['y text' if loc=='above' else 'x text']=np.linspace(
                                (axlims['y' if loc=='above' else 'x']['min']+axlims['y' if loc=='above' else 'x']['len']*0.5)+offy,
                                axlims['y' if loc=='above' else 'x']['max']-offy,
                                len(df1_))
        df1_.apply(lambda x: ax.annotate(s=x[coltext], 
                                        xytext=(x['x text'],x['y text']), 
                                        xy=(x[colx],x[coly]), 
                                        va='center',
                                        ha='center',
                                         **kws,
                                        ),axis=1)
    return ax

# variance
from matplotlib.patches import Ellipse
import matplotlib.transforms as transforms
def confidence_ellipse(x, y, ax, n_std=3.0, facecolor='none', **kwargs):
    """
    Create a plot of the covariance confidence ellipse of *x* and *y*.

    Parameters:
    ----------
    x, y : array-like, shape (n, )
        Input data.

    ax : matplotlib.axes.Axes
        The axes object to draw the ellipse into.

    n_std : float
        The number of standard deviations to determine the ellipse's radiuses.

    **kwargs
        Forwarded to `~matplotlib.patches.Ellipse`

    Returns
    -------
    matplotlib.patches.Ellipse
    
    References
    ----------
    https://matplotlib.org/3.5.0/gallery/statistics/confidence_ellipse.html
    """
    if x.size != y.size:
        raise ValueError("x and y must be the same size")

    cov = np.cov(x, y)
    pearson = cov[0, 1]/np.sqrt(cov[0, 0] * cov[1, 1])
    # Using a special case to obtain the eigenvalues of this
    # two-dimensionl dataset.
    ell_radius_x = np.sqrt(1 + pearson)
    ell_radius_y = np.sqrt(1 - pearson)
    ellipse = Ellipse((0, 0), width=ell_radius_x * 2, height=ell_radius_y * 2,
                      facecolor=facecolor, **kwargs)

    # Calculating the stdandard deviation of x from
    # the squareroot of the variance and multiplying
    # with the given number of standard deviations.
    scale_x = np.sqrt(cov[0, 0]) * n_std
    mean_x = np.mean(x)

    # calculating the stdandard deviation of y ...
    scale_y = np.sqrt(cov[1, 1]) * n_std
    mean_y = np.mean(y)

    transf = transforms.Affine2D() \
        .rotate_deg(45) \
        .scale(scale_x, scale_y) \
        .translate(mean_x, mean_y)

    ellipse.set_transform(transf + ax.transData)
    return ax.add_patch(ellipse)

## heatmaps
def show_box(
    ax: plt.Axes,
    xy: tuple,
    width: float,
    height: float,
    fill: str=None, 
    alpha: float=1,
    lw: float=1.1,
    ec: str='k',
    clip_on: bool=False,
    scale_width: float=1,
    scale_height: float=1,
    xoff: float=0,
    yoff: float=0,
    **kws,
    ) -> plt.Axes:
    """Highlight sections of a plot e.g. heatmap by drawing boxes.

    Args:
        xy (tuple): position of left, bottom corner of the box.
        width (float): width.
        height (float): height.
        ax (plt.Axes, optional): `plt.Axes` object. Defaults to None.
        fill (str, optional): fill the box with color. Defaults to None.
        alpha (float, optional): alpha of color. Defaults to 1.
        lw (float, optional): line width. Defaults to 1.1.
        ec (str, optional): edge color. Defaults to 'k'.
        clip_on (bool, optional): clip the boxes by the axis limit. Defaults to False.
        scale_width (float, optional): scale width. Defaults to 1.
        scale_height (float, optional): scale height. Defaults to 1.
        xoff (float, optional): x-offset. Defaults to 0.
        yoff (float, optional): y-offset. Defaults to 0.

    Keyword Args:
        kws: parameters provided to the `Rectangle` function. 

    Returns:
        plt.Axes: `plt.Axes` object.
    """
    from matplotlib.patches import Rectangle
    return ax.add_patch(Rectangle(
                            xy=[xy[0]+xoff,xy[1]+yoff], 
                            width=width*scale_width, height=height*scale_height, 
                            fill=fill, alpha=alpha,
                            lw=lw,ec=ec,clip_on=clip_on,
                            **kws,
                            ))

def annot_confusion_matrix(
    df_: pd.DataFrame,
    ax: plt.Axes=None,
    off: float=0.5
    ) -> plt.Axes:
    """Annotate a confusion matrix.

    Args:
        df_ (pd.DataFrame): input data.
        ax (plt.Axes, optional): `plt.Axes` object. Defaults to None.
        off (float, optional): offset. Defaults to 0.5.

    Returns:
        plt.Axes: `plt.Axes` object.
    """
    from roux.stat.binary import get_stats_confusion_matrix
    df1=get_stats_confusion_matrix(df_)
    df2=pd.DataFrame({
                    'TP': [0,0],
                    'TN': [1,1],
                    'FP': [0,1],
                    'FN': [1,0],
                    'TPR':[0,2],
                    'TNR': [1,2],
                    'PPV': [2,0],
                    'NPV': [2,1],
                    'FPR': [1,3],
                    'FNR': [0,3],
                    'FDR': [3,0],
                    'ACC': [2,2],
                    },
                     index=['x','y']).T
    df2.index.name='variable'
    df2=df2.reset_index()
    df3=df1.merge(df2,
              on='variable',
              how='inner',
              validate="1:1")
    
    _=df3.loc[(df3['variable'].isin(['TP','TN','FP','FN'])),:].apply(lambda x: ax.text(x['x']+off,
                                                                                       x['y']+(off*2),
    #                               f"{x['variable']}\n{x['value']:.0f}",
                                  x['variable'],
    #                               f"({x['T|F']+x['P|N']})",
                                ha='center',va='bottom',
                               ),axis=1)
    _=df3.loc[~(df3['variable'].isin(['TP','TN','FP','FN'])),:].apply(lambda x: ax.text(x['x']+off,
                                                                                        x['y']+(off*2),
                                  f"{x['variable']}\n{x['value']:.2f}",
    #                               f"({x['T|F']+x['P|N']})",
                                ha='center',va='bottom',
                               ),axis=1)
    return ax


def get_logo_ax(
    ax: plt.Axes,
    size: float=0.5,
    bbox_to_anchor: list=None,
    loc: str=1,
    axes_kwargs: dict={'zorder':-1},
    ) -> plt.Axes:
    """Get `plt.Axes` for placing the logo.

    Args:
        ax (plt.Axes): `plt.Axes` object.
        size (float, optional): size of the subplot. Defaults to 0.5.
        bbox_to_anchor (list, optional): location. Defaults to None.
        loc (str, optional): location. Defaults to 1.
        axes_kwargs (_type_, optional): parameters provided to `inset_axes`. Defaults to {'zorder':-1}.

    Returns:
        plt.Axes: `plt.Axes` object.
    """
    from mpl_toolkits.axes_grid1.inset_locator import inset_axes
    width, height,aspect_ratio=get_subplot_dimentions(ax)
    axins = inset_axes(ax, 
                       width=size, height=size,
                       bbox_to_anchor=[1,1,0,size/(height)] if bbox_to_anchor is None else bbox_to_anchor,
                       bbox_transform=ax.transAxes, 
                       loc=loc, 
                       borderpad=0,
                      axes_kwargs=axes_kwargs)
    return axins

def set_logo(
    imp: str,
    ax: plt.Axes,
    size: float=0.5,
    bbox_to_anchor: list=None,
    loc: str=1,
    axes_kwargs: dict={'zorder':-1},
    params_imshow: dict={'aspect':'auto','alpha':1,
    #                             'zorder':1,
    'interpolation':'catrom'},
    test: bool=False,
    force: bool=False
    ) -> plt.Axes:
    """Set logo.

    Args:
        imp (str): path to the logo file.
        ax (plt.Axes): `plt.Axes` object.
        size (float, optional): size of the subplot. Defaults to 0.5.
        bbox_to_anchor (list, optional): location. Defaults to None.
        loc (str, optional): location. Defaults to 1.
        axes_kwargs (_type_, optional): parameters provided to `inset_axes`. Defaults to {'zorder':-1}.
        params_imshow (_type_, optional): parameters provided to the `imshow` function. Defaults to {'aspect':'auto','alpha':1, 'interpolation':'catrom'}.
        test (bool, optional): test mode. Defaults to False.
        force (bool, optional): overwrite file. Defaults to False.

    Returns:
        plt.Axes: `plt.Axes` object.
    """
    from roux.lib.figs.convert import vector2raster
    if isinstance(imp,str):
        if splitext(imp)[1]=='.svg':
            pngp=vector2raster(imp,force=force)
        else:
            pngp=imp
        if not exists(pngp):
            logging.error(f'{pngp} not found')
            return
        im = plt.imread(pngp)
    elif isinstance(imp,np.ndarray):
        im = imp
    else:
        loggin.warning('imp should be path or image')
        return
    axins=get_logo_ax(ax,size=size,bbox_to_anchor=bbox_to_anchor,loc=loc,
             axes_kwargs=axes_kwargs,)
    axins.imshow(im, **params_imshow)
    if not test:
        axins.set(**{'xticks':[],'yticks':[],'xlabel':'','ylabel':''})
        axins.margins(0)    
        axins.axis('off')    
        axins.set_axis_off()
    else:
        print(width, height,aspect_ratio,size/(height*2))
    return axins

## color
def color_ax(
    ax: plt.Axes,
    c: str,
    linewidth: float=None
    ) -> plt.Axes:
    """Color border of `plt.Axes`.

    Args:
        ax (plt.Axes): `plt.Axes` object.
        c (str): color.
        linewidth (float, optional): line width. Defaults to None.

    Returns:
        plt.Axes: `plt.Axes` object.
    """
    plt.setp(ax.spines.values(), color=c)
    if not linewidth is None:
        plt.setp(ax.spines.values(), linewidth=linewidth)
    plt.setp([ax.get_xticklines(), ax.get_yticklines()], color=c)
    return ax

## stats
def annot_n_legend(
    ax,
    df1: pd.DataFrame,
    colid: str,
    colgroup: str,
    **kws,
    ):
    from roux.viz.ax_ import rename_legends
    replaces={str(k):str(k)+'\n'+f'(n={v})' for k,v in df1.groupby(colgroup)[colid].nunique().items()}
    return rename_legends(ax,
                  replaces=replaces,
                  **kws)


import pandas as pd
import numpy as np
import scipy as sc
import logging
from roux.lib.sys import info

from scipy.stats import spearmanr,pearsonr
def get_spearmanr(x: np.array,y: np.array) -> tuple:
    """Get Spearman correlation coefficient.

    Args:
        x (np.array): x vector.
        y (np.array): y vector.

    Returns:
        tuple: rs, p-value
    """
    t=sc.stats.spearmanr(x,y,nan_policy='omit')
    return t.correlation,float(t.pvalue)

def get_pearsonr(x: np.array,y: np.array) -> tuple:
    """Get Pearson correlation coefficient.

    Args:
        x (np.array): x vector.
        y (np.array): y vector.

    Returns:
        tuple: rs, p-value
    """
    return sc.stats.pearsonr(x,y)

def get_corr_bootstrapped(x: np.array,
                          y: np.array,
                          method='spearman',
                          ci_type='max',
                          cv:int=5,
                          random_state=1,
                          verbose=False,
                         ) -> tuple:
    """Get correlations after bootstraping.

    Args:
        x (np.array): x vector.
        y (np.array): y vector.
        method (str, optional): method name. Defaults to 'spearman'.
        ci_type (str, optional): confidence interval type. Defaults to 'max'.
        cv (int, optional): number of bootstraps. Defaults to 5.
        random_state (int, optional): random state. Defaults to 1.

    Returns:
        tuple: mean correlation coefficient, confidence interval
    """
    from roux.stat.classify import get_cvsplits
    from roux.stat.variance import get_ci
    cv2xy=get_cvsplits(x,y,cv=cv,outtest=False,random_state=1)
    rs=[globals()[f"get_{method}r"](**cv2xy[k])[0] for k in cv2xy]
    if verbose: info(cv,ci_type)
    return np.mean(rs), get_ci(rs,ci_type=ci_type)

def corr_to_str(method: str,
                r: float,p: float,
                fmt='<',n=True, ci=None,ci_type=None, magnitide=True) -> str:
    """Correlation to string

    Args:
        method (str): method name.
        r (float): correlation coefficient.
        p (float): p-value
        fmt (str, optional): format of the p-value. Defaults to '<'.
        n (bool, optional): sample size. Defaults to True.
        ci (_type_, optional): confidence interval. Defaults to None.
        ci_type (_type_, optional): confidence interval type. Defaults to None.
        magnitide (bool, optional): show magnitude of the sample size. Defaults to True.

    Returns:
        str: string with the correation stats. 
    """
    from roux.viz.annot import pval2annot
    from roux.lib.str import num2str
    s0=f"$r_{method[0]}$={r:.2f}"
    if not ci is None:
        s0+=f"$\pm${ci:.2f}{ci_type if ci_type!='max' else ''}"
    s0+=f"\n{pval2annot(p,fmt='<',linebreak=False, alpha=0.05)}"+('' if not n else f"\nn="+num2str(num=n,magnitude=False))
    return s0

def get_corr(
    x: np.array,y: np.array,
    method='spearman',
    bootstrapped=False,
    ci_type='max',
    magnitide=True,
    outstr=False,
    **kws):
    """Correlation between vectors (wrapper).

    Args:
        x (np.array): x.
        y (np.array): y.
        method (str, optional): method name. Defaults to 'spearman'.
        bootstrapped (bool, optional): bootstraping. Defaults to False.
        ci_type (str, optional): confidence interval type. Defaults to 'max'.
        magnitide (bool, optional): show magnitude. Defaults to True.
        outstr (bool, optional): output as string. Defaults to False.
    
    Keyword arguments:
        kws: parameters provided to `get_corr_bootstrapped` function.

    """
    n=len(x)
    if bootstrapped:
        r,ci=get_corr_bootstrapped(x,y,method=method,ci_type=ci_type,**kws)
        _,p=globals()[f"get_{method}r"](x, y)
        if not outstr:
            return r,p,ci,n
        else:
            return corr_to_str(method,r,p,n=n, ci=ci,ci_type=ci_type, magnitide=magnitide),r
    else:
        r,p=globals()[f"get_{method}r"](x, y)
        if not outstr:
            return r,p,n
        else:
            return corr_to_str(method,r,p,n=n, ci=None,ci_type=None, magnitide=magnitide),r

def get_corrs(
    df1,
    method,
    cols,
    cols_with=[],
    coff_inflation_min=None,
    test=False,
    **kws
    ):
    """Correlate columns of a dataframes.

    Args:
        df1 (DataFrame): input dataframe.
        method (str): method of correlation `spearman` or `pearson`.        
        cols (str): columns.
        cols_with (str): columns to correlate with i.e. variable2.

    Keyword arguments:
        kws: parameters provided to `get_corr` function.

    Returns:
        DataFrame: output dataframe.
        
    TODOs:
        0. use `lib.set.get_pairs` to get the combinations.
        1. Provide 2D array to `scipy.stats.spearmanr`?
        2. Add parallel processing through `fast` parameter.
    """
    import itertools
    from roux.stat.diff import get_q
    # check inflation/over-representations
    from roux.lib.df import check_inflation
    ds_=check_inflation(df1,subset=cols+cols_with).loc[lambda x: x>=(50 if coff_inflation_min is None else coff_inflation_min)]
    info(ds_)
    if not coff_inflation_min is None:
        # remove inflated
        cols=[c for c in cols if not c in ds_.index.tolist()]
        cols_with=[c for c in cols_with if not c in ds_.index.tolist()]
    if len(cols_with)==0:
        o1=itertools.combinations(cols,2)
    else:
        o1=itertools.product(cols,cols_with)
    # remove inf
    df1=df1.loc[:,np.unique(cols+cols_with)].replace([np.inf, -np.inf], np.nan)
    
    df0=pd.DataFrame(o1,columns=['variable1','variable2'])
    df0=df0.loc[(df0['variable1']!=df0['variable2']),:]
    if test:
        info(df0)
    df2=df0.groupby(['variable1','variable2']).progress_apply(lambda df: get_corr(x=df1[df.name[0]],
                                                                                  y=df1[df.name[1]],
                                                                                  method=method,
                                                                                  **kws)).apply(pd.Series)
    df2.columns=[f"$r_{method[0]}$",'P','n']
    df2=(df2
        .reset_index()
        .log.dropna(subset=['P'])
        .groupby(['variable1']+(['variable2'] if len(cols_with)==0 else []),as_index=False).apply(lambda df: get_q(df,'P')).reset_index(drop=True)
        .sort_values(['Q',f"$r_{method[0]}$"],ascending=[True,False])
        )
    return df2

## partial 
def get_partial_corrs(
    df: pd.DataFrame,
    xs: list,
    ys: list, 
    method='spearman',
    splits=5
    ) -> pd.DataFrame:
    """Get partial correlations.

    Args:
        df (DataFrame): input dataframe.
        xs (list): columns used as x variables.
        ys (list): columns used as y variables.
        method (str, optional): method name. Defaults to 'spearman'.
        splits (int, optional): number of splits. Defaults to 5.

    Returns:
        DataFrame: output dataframe.
    """
    import pingouin as pg
    import itertools
    chunks=np.array_split(df.sample(frac=1,random_state=88),splits)
    dn2df={}
    for chunki,chunk in enumerate(chunks):
        dn2df_={}
        for x,y in list(itertools.product(xs,ys)):
            if (x if isinstance(x,str) else x[0])!=(y if isinstance(y,str) else y[0]): 
                params=dict(
                        x=(x if isinstance(x,str) else x[0]),x_covar=(None if isinstance(x,str) else x[1:]), 
                        y=(y if isinstance(y,str) else y[0]),y_covar=(None if isinstance(y,str) else y[1:]), )
                label=str(params)#+(x if isinstance(x,str) else f"{x[0]} (corrected for {' '.join(x[1:])})")+" versus "+(y if isinstance(y,str) else f"{y[0]} (corrected for {' '.join(y[1:])})")
                dn2df_[label]=pg.partial_corr(data=chunk, 
                                        tail='two-sided', method=method,
                                         **params)
#                 print(dn2df_[label])
#                 print(params)
                for k in params:
                    dn2df_[label].loc[method,k]=params[k] if isinstance(params[k],str) else str(params[k])
        dn2df[chunki]=pd.concat(dn2df_,axis=0)
    df1=pd.concat(dn2df,axis=0)
    df1.index.names=['chunk #','correlation name','correlation method']
    for c in ['x','y']:
        if f"{c}_covar" in df1:
            df1[f"{c}_covar"]=df1[f"{c}_covar"].apply(eval)
    return df1.reset_index()

def check_collinearity(
    df1,
    threshold=0.7,
    colvalue='$r_s$',
    cols_variable=['variable1','variable2'],
    coff_pval=0.05,
    )-> pd.Series:
    """Check collinearity.

    Args:
        df1 (DataFrame): input dataframe.
        threshold (float): minimum threshold for the colinearity.

    Returns:
        DataFrame: output dataframe.
    
    TODOs:
        1. Calculate variance inflation factor (VIF).
    """
    cols=df1.columns.tolist()
    df2=get_corrs(df1=df1,
        method='spearman',
        cols=cols,
        cols_with=cols,
        coff_inflation_min=50,
        )
    df2=df2.loc[(df2['P']<0.05),:]

    df2['is collinear']=df2[colvalue].abs().apply(lambda x: x>=threshold)
    perc=(df2['is collinear'].sum()/len(df2))*100
    logging.info(f"% collinear vars: {perc} ({df2['is collinear'].sum()}/{len(df1.columns)})")
    if perc==0:
        logging.info(f"max corr={df2[colvalue].max()}")
        return
    df2=df2.loc[(df2['is collinear']),:]
    from roux.stat.network import get_subgraphs
    df3=get_subgraphs(df2.loc[df2['is collinear'],:],cols_variable[0],cols_variable[1])
    df3=df3.groupby('subnetwork name').agg({'node name':list}).reset_index()
    return (df3
            .groupby('subnetwork name')
            .progress_apply(lambda df: df2.apply(lambda x: x[colvalue] if len(set([x[cols_variable[0]],x[cols_variable[1]]]) - set(df['node name'].tolist()[0]))==0 else np.nan,axis=1).min())
            .sort_values(ascending=False)
           )    

def pairwise_chi2(
    df1: pd.DataFrame,
    cols_values: list
    ) -> pd.DataFrame:
    """Pairwise chi2 test.

    Args:
        df1 (DataFrame): pd.DataFrame
        cols_values (list): list of columns.

    Returns:
        DataFrame: output dataframe.
        
    TODOs:
        0. use `lib.set.get_pairs` to get the combinations.
    """
    import itertools
    d1={}
    for cols in list(itertools.combinations(cols_values,2)):
        _,d1[cols],_,_=sc.stats.chi2_contingency(pd.crosstab(df1[cols[0]],df1[cols[1]]))

    df2=pd.Series(d1).to_frame('P')
    df2.index.names=['value1','value2']
    df2=df2.reset_index()
    return df2
